# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""
pymwp: implementation of MWP analysis on C code in Python.
"""

__title__ = "pymwp"
__author__ = "Clément Aubert, Thomas Rubiano, Neea Rusch, Thomas Seiller"
__license__ = "GPL-3.0-or-later"

__notice__ = (
    f"{__title__} Copyright (c) 2020-2024 {__author__}. This program comes "
    f"with ABSOLUTELY NO WARRANTY; for details type `{__title__} --license W`."
    f"This is free software, and you are welcome to redistribute it under "
    f"certain conditions; type `{__title__} --license C` for details.")

__warranty__ = (
    f"{__title__} is distributed in the hope that it will be useful, but "
    f"WITHOUT ANY WARRANTY; without even the implied warranty of "
    f"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU "
    f"General Public License for more details "
    f"https://www.gnu.org/licenses/gpl-3.0.html.")

__conditions__ = (
    "Permissions of this strong copyleft license are conditioned on making "
    "available complete source code of licensed works and modifications, "
    "which include larger works using a licensed work, under the same "
    "license. Copyright and license notices must be preserved. Contributors "
    f"provide an express grant of patent rights. See the GNU General Public "
    f"License for more details https://www.gnu.org/licenses/gpl-3.0.html.")

# flake8: noqa: F401,E402
from pymwp.version import __version__
from pymwp.constants import *  # import all types
from pymwp.parser import Parser
from pymwp.choice import Choices
from pymwp.syntax import Coverage, Variables, FindLoops
from pymwp.monomial import Monomial
from pymwp.delta_graphs import DeltaGraph
from pymwp.polynomial import Polynomial
from pymwp.relation import Relation
from pymwp.relation_list import RelationList
from pymwp.bound import Bound, MwpBound
from pymwp.result import Result, FuncResult, FuncLoops, LoopResult, VResult
from pymwp.analysis import Analysis, LoopAnalysis



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
from abc import abstractmethod
from collections import Counter
from copy import deepcopy
from typing import List, Callable, Tuple, Optional, Union

# noinspection PyPep8Naming
from .parser import Parser as pr, NodeHandler

logger = logging.getLogger(__name__)


class SyntaxUtils:
    """Various helpful syntax-related utility methods."""

    @staticmethod
    def array_name(node: pr.ArrayRef) -> str:
        """Find array identifier.

        Arguments:
            node (pr.ArrayRef): Array AST node.

        Returns:
            Find array name from node.
        """
        name_node = node.name
        while not isinstance(name_node, pr.ID):
            name_node = name_node.name
        return name_node.name

    @staticmethod
    def init_vars(node: Union[pr.Assignment, pr.DeclList, pr.ExprList]) \
            -> Tuple[List[str], List[str]]:
        """Find and group variables in an init-block.

        Looks for declarations/iterators `int i=…,…` on left
        and "source" variables `…=y` on right.

        Arguments:
            node: Node to inspect.

        Returns:
            Discovered variable lists of declarations and sources.
        """

        def att(lst, attr):
            return [getattr(e, attr) for e in lst]

        def names(lst):
            return [e.name for e in lst if isinstance(e, (pr.ID, pr.Decl))]

        # int i=0,… one or more declarations
        if isinstance(node, pr.DeclList):
            return names(node.decls), names(att(node.decls, 'init'))

        # i=0, j=x,… one or more assignments
        exp = node.exprs if hasattr(node, 'exprs') else [node]
        return names(att(exp, 'lvalue')), names(att(exp, 'rvalue'))

    @staticmethod
    def rm_child(node: pr.Node, attr: str, child: pr.Node) -> Callable:
        """Construct a callable function to remove a child node.

        Using a callable allows performing the actual removal later.

        Arguments:
            node (pr.Node): Parent node.
            attr (str): Parent's attribute that contains child, e.g., stmt.
            child (pr.Node): Child node.

        Returns:
            A callable function.
        """
        return lambda: getattr(node, attr).remove(child) \
            if child in getattr(node, attr) else None

    @staticmethod
    def rm_attr(node: pr.Node, attr: str) -> Callable:
        """Construct a callable function to clear an attribute.

        Arguments:
            node (pr.Node): AST node.
            attr (str): Attribute to remove.

        Returns:
            A callable function.
        """
        return lambda: setattr(node, attr, pr.EmptyStatement())

    @staticmethod
    def _fmt(idx: int, count: int, desc: str) -> str:
        """Formatter for displaying unsupported nodes.

        Arguments:
            idx (int): Ranked order (1., 2., 3.…).
            count (int): Number of occurrences.
            desc (str): Node description.

        Returns:
            Formatted string expression for display.
        """
        order = f"{(str(idx) + '.'):<4}"
        times = f" {(str(count) + 'x ')}" if count > 1 else ' '
        return f"{order}{times}{desc}"

    @staticmethod
    def unsupported(omits: List[str]) -> None:
        """Display unsupported nodes as an ordered list.

        Arguments:
            omits (List[str]): List of unsupported AST nodes.
        """
        n = len(omits)
        if n == 1:
            logger.debug(f'Unsupported syntax: {omits[0]}')
        elif n > 1:
            codes = [(cnt, code) for (code, cnt) in Counter(omits).items()]
            codes = sorted(codes, key=lambda x: (-x[0], x[1]))
            logger.debug(f'Unsupported syntax {n}x, {len(codes)} unique')
            for i, cv in enumerate(codes):
                logger.debug(SyntaxUtils._fmt(i, *cv))

    @staticmethod
    def print_mod(node: pr.Node) -> pr.Node:
        """Prepare AST node for display as string.

        For example, for long blocks the body statement is removed.
        The original node is never modified; if some edit is applied, it is
        always applied to a copy of the AST node.

        Arguments:
            node (pr.Node): AST node.

        Returns:
            AST node conditionally formatted for display.
        """
        if isinstance(node, pr.ArrayRef):
            node = deepcopy(node)
            node.name = pr.ID(SyntaxUtils.array_name(node))
            node.subscript = pr.Constant('String', '…')
        elif hasattr(node, 'args') and hasattr(node.args, 'exprs'):
            node = deepcopy(node)
            node.args = pr.ExprList([pr.Constant('String', '…')])
        elif hasattr(node, 'stmt'):
            node = deepcopy(node)
            node.stmt = None
        return node


# noinspection PyPep8Naming
class BaseAnalysis(NodeHandler):
    """Base implementation for AST analysis."""

    PLUS, MINUS, MULT = '+', '-', '*'
    NEG, SIZEOF = '!', 'sizeof'

    BIN_OPS = {PLUS, MINUS, MULT}
    """Supported binary operators."""

    SIGN = {'+', '-'}
    PREFIX = {'++', '--'}
    INC, DEC = {'p++', '++'}, {'p--', '--'}
    INC_DEC = INC | DEC
    U_OPS = INC | DEC | {PLUS, MINUS, NEG, SIZEOF}
    """Supported unary operators."""

    @abstractmethod
    def handler(self, node: pr.Node, *args, **kwargs) \
            -> None:  # pragma: no cover
        """Handler for AST nodes that meet some abstract criteria."""
        pass

    def node_handler(self, node: pr.Node) -> Callable:
        # Maps node type to method to call.
        t_name = type(node).__name__
        return getattr(self, t_name) if hasattr(self, t_name) \
            else self.handler

    def recurse(self, node: pr.Node, *args, **kwargs) -> None:
        """Traverse AST nodes."""
        return (self.FCall(node, *args, **kwargs)
                if isinstance(node, pr.FuncCall) else
                self.node_handler(node)(node, *args, **kwargs))

    def _recurse_attr(self, node: pr.Node, attr: str, *args, **kwargs):
        """Analyze node's child at attribute, if exists."""
        if hasattr(node, attr) and getattr(node, attr):
            self.recurse(getattr(node, attr), *args, **kwargs)

    def _iter_attr(self, node: pr.Node, attr: str, *args, **kwargs):
        """Analyze node's children at attribute, if any exist."""
        if hasattr(node, attr) and getattr(node, attr):
            for n in getattr(node, attr):
                self.recurse(n, *args, **kwargs)

    def Assert(self, node: pr.FuncCall, *args, **kwargs):
        pass

    def Assume(self, node: pr.FuncCall, *args, **kwargs):
        pass

    def FCall(self, node: pr.FuncCall, *args, **kwargs):
        # Routes function call to the correct handler
        if isinstance(node.name, pr.ID):
            if node.name.name == 'assert':
                return self.Assert(node, *args, **kwargs)
            if node.name.name == 'assume':
                return self.Assume(node, *args, **kwargs)
        return self.FuncCall(node, *args, **kwargs)

    def Case(self, node: pr.Case, *args, **kwargs):
        self._iter_attr(node, 'stmts', *args, **kwargs)

    def Compound(self, node: pr.Compound, *args, **kwargs):
        self._iter_attr(node, 'block_items', *args, **kwargs)

    def DeclList(self, node: pr.DeclList, *args, **kwargs):
        self._iter_attr(node, 'decls', *args, **kwargs)

    def Default(self, node: pr.Default, *args, **kwargs):
        self._iter_attr(node, 'stmts', *args, **kwargs)

    def ExprList(self, node: pr.ExprList, *args, **kwargs):
        self._iter_attr(node, 'exprs', *args, **kwargs)

    def ParamList(self, node: pr.ParamList, *args, **kwargs):
        self._iter_attr(node, 'params', *args, **kwargs)


class Coverage(BaseAnalysis):
    """Simple analysis of AST syntax to determine if AST contains
    only supported language features.

    Attributes:
        node(pr.Node): AST node.
        omit(List[str]): List of unsupported commands.
        clear_list(List[Callable]): Functions to clear unsupported syntax.
    """

    def __init__(self, node: pr.Node):
        self.node = node
        self.omit = []
        self.clear_list = []
        self.recurse(self.node)

    @property
    def full(self) -> bool:
        """True if entire syntax tree is fully covered by analysis."""
        return len(self.omit) == 0

    def report(self) -> Coverage:
        """Display syntax coverage for inspected AST node."""
        SyntaxUtils.unsupported(self.omit)
        return self

    def ast_mod(self) -> Coverage:
        """Removes unsupported AST nodes in place."""
        n = len(self.omit)
        assert (n == len(self.clear_list))
        [callable_() for callable_ in self.clear_list]
        self.omit, self.clear_list = [], []
        logger.debug(f"Removed unsupported syntax: {n} node(s)")
        return self

    @staticmethod
    def loop_compat(node: pr.For) -> Tuple[bool, Optional[str]]:
        r"""Check if C-language for loop is compatible with an "mwp-loop".

        The mwp-loop has form $\text{loop} \; \texttt{X} \; \{ \texttt{C} \}$.
        Try to identify if C-language `for` loop has a similar form,
        _repeat X times command C_. The variable X is not allowed to
        occur in the body C of the loop.

        Arguments:
            node: AST node to inspect.

        Returns:
            A tuple where the first item is a compatibility result: True if
                for loop is mwp-loop compatible, otherwise False. The second
                is the name of iteration guard variable X, possibly `None`.
        """
        loop_x, body = Variables.loop_guard(node)
        if len(loop_x) != 1:  # exactly one guard variable
            logger.debug(f"Unknown loop guard variable in {loop_x}")
            return False, None
        x_var = loop_x[0]
        if x_var in body:
            logger.warning(f"Variable {x_var} occurs in loop body")
            return False, None
        return True, x_var

    def handler(self, node: pr.Node, *args, **kwargs):
        """Make a list of uncovered nodes."""
        # add to clear list
        self.clear_list.append(kwargs['clear'])  # should always exist
        # display edit, then add to list of issues to display
        node = SyntaxUtils.print_mod(node)
        self.omit.append(pr.to_c(node, compact=True))

    def recurse(self, node: pr.Node, *args, **kwargs):
        if isinstance(node, (pr.TernaryOp, pr.ArrayRef, pr.Switch)):
            self.handler(node, *args, **kwargs)
        else:
            super().recurse(node, *args, **kwargs)

    def _iter_attr(self, node: pr.Node, attr: str, *args, **kwargs):
        if hasattr(node, attr) and getattr(node, attr):
            for n in getattr(node, attr):
                cl = SyntaxUtils.rm_child(node, attr, n)
                self.recurse(n, *args, **{**kwargs, 'clear': cl})

    def FuncCall(self, node: pr.FuncCall, *args, **kwargs):
        self.handler(node, *args, **kwargs)

    def Assignment(self, node: pr.Assignment, *args, **kwargs):
        allow = pr.BinaryOp, pr.Constant, pr.ID, pr.UnaryOp
        right = (node.rvalue.expr if isinstance(node.rvalue, pr.Cast)
                 else node.rvalue)
        if not (node.op == "=" and isinstance(node.lvalue, pr.ID) and
                isinstance(right, allow)):
            self.handler(node, *args, **kwargs)
        else:
            self.recurse(node.lvalue, *args, **kwargs)
            self.recurse(right, *args, **kwargs)

    def BinaryOp(self, node: pr.BinaryOp, *args, **kwargs):
        left, right = node.left, node.right
        lf = left.expr if isinstance(left, pr.Cast) else left
        rt = right.expr if isinstance(right, pr.Cast) else right
        allow = pr.Constant, pr.ID
        if not (node.op in self.BIN_OPS and isinstance(lf, allow) and
                isinstance(rt, allow)):
            self.handler(node, *args, **kwargs)

    def Cast(self, node: pr.Cast, *args, **kwargs):
        self._recurse_attr(node, 'expr', *args, **kwargs)

    def Decl(self, node: pr.Decl, *args, **kwargs):
        init = node.init if hasattr(node, 'init') else None
        type_ = node.type if hasattr(node, 'type') else None
        if not (isinstance(type_, pr.TypeDecl) and init is None):
            self.handler(node, *args, **kwargs)

    def DoWhile(self, node: pr.DoWhile, *args, **kwargs):
        self._recurse_attr(node, 'stmt', *args, **kwargs)

    def For(self, node: pr.For, *args, **kwargs):
        if not self.loop_compat(node)[0]:
            self.handler(node, *args, **kwargs)
        else:
            self._recurse_attr(node, 'stmt', *args, **kwargs)

    def FuncDef(self, node: pr.FuncDef, *args, **kwargs):
        self._recurse_attr(node.decl.type, 'args', *args, **kwargs)
        self._recurse_attr(node, 'body', *args, **kwargs)

    def If(self, node: pr.If, *args, **kwargs):
        t_kwargs = {**kwargs, 'clear': SyntaxUtils.rm_attr(node, 'iftrue')}
        self._recurse_attr(node, 'iftrue', *args, **t_kwargs)
        e_kwargs = {**kwargs, 'clear': SyntaxUtils.rm_attr(node, 'iffalse')}
        self._recurse_attr(node, 'iffalse', *args, **e_kwargs)

    def Return(self, node: pr.Return, *args, **kwargs):
        self._recurse_attr(node, 'expr', *args, **kwargs)

    def UnaryOp(self, node: pr.UnaryOp, *args, **kwargs):
        if not (node.op in self.U_OPS and isinstance(
                node.expr, (pr.ID, pr.Constant, pr.Cast, pr.UnaryOp))):
            self.handler(node, *args, **kwargs)
        else:
            self._recurse_attr(node, 'expr', *args, **kwargs)

    def While(self, node: pr.While, *args, **kwargs):
        self._recurse_attr(node, 'stmt', *args, **kwargs)


class FindLoops(BaseAnalysis):
    """Finds all loop nodes in an AST.

    Attributes:
        loops (List[pr.LoopT]): Loop nodes.
    """

    def __init__(self, node: pr.Node):
        self.loops: List[pr.LoopT] = []
        self.recurse(node)

    def handler(self, node: pr.LoopT, *args, **kwargs) -> None:
        """Make a (flat) list of the discovered loops."""
        self.loops.append(node)

    def DoWhile(self, node: pr.DoWhile, *args, **kwargs):
        self.handler(node, *args, **kwargs)
        self._recurse_attr(node, 'stmt', *args, **kwargs)

    def For(self, node: pr.For, *args, **kwargs):
        if Coverage.loop_compat(node)[0]:
            self.handler(node, *args, **kwargs)
        self._recurse_attr(node, 'stmt', *args, **kwargs)

    def FuncDef(self, node: pr.FuncDef, *args, **kwargs):
        self._recurse_attr(node, 'body', *args, **kwargs)

    def If(self, node: pr.If, *args, **kwargs):
        self._recurse_attr(node, 'iftrue', *args, **kwargs)
        self._recurse_attr(node, 'iffalse', *args, **kwargs)

    def Switch(self, node: pr.Switch, *args, **kwargs):
        self._recurse_attr(node, 'stmt', *args, **kwargs)

    def While(self, node: pr.While, *args, **kwargs):
        self.handler(node, *args, **kwargs)
        self._recurse_attr(node, 'stmt', *args, **kwargs)


class Variables(BaseAnalysis):
    """Find variables in an AST.

    Attributes:
        vars (List[str]): List of variables.
    """

    RESERVED = ['true', 'false']
    """List of reserved names that are not variables;
    see [issue #150](https://github.com/statycc/pymwp/issues/150)."""

    def __init__(self, *nodes: pr.Node):
        self.vars = []
        [self.recurse(node) for node in nodes]
        self.vars = sorted(self.vars)

    @staticmethod
    def loop_guard(node: pr.For) -> Tuple[List[str], List[str]]:
        """Find variables in a for loop.

        Arguments:
            node (pr.For): A for-loop AST node.

        Returns:
            Two lists of variables: `(loop_guard, body_variables)`.
        """
        iters, srcs = SyntaxUtils.init_vars(node.init)
        conds = Variables(node.cond).vars
        nxt = Variables(node.next).vars
        iters = list(set(iters) | set(nxt))
        body = Variables(node.stmt).vars

        loop_vars = set(conds) | set(srcs)
        loop_x = list(loop_vars - set(iters))

        info = [('guard', loop_x), ('body', body)]
        info = [f"{lbl}: {' '.join(v) or '?'}" for lbl, v in info]
        logger.debug(f"for-loop {', '.join(info)}")
        return loop_x, body

    def handler(self, node: pr.Node, *args, **kwargs):
        """Record the name of a discovered variable."""
        if hasattr(node, 'name') and node.name:
            if node.name not in self.vars:
                self.vars.append(node.name)

    def Assignment(self, node: pr.Assignment, *args, **kwargs):
        self._recurse_attr(node, 'lvalue', *args, **kwargs)
        self._recurse_attr(node, 'rvalue', *args, **kwargs)

    def BinaryOp(self, node: pr.BinaryOp, *args, **kwargs):
        self._recurse_attr(node, 'left', *args, **kwargs)
        self._recurse_attr(node, 'right', *args, **kwargs)

    def Cast(self, node: pr.Cast, *args, **kwargs):
        self._recurse_attr(node, 'expr', *args, **kwargs)

    def Decl(self, node: pr.Decl, *args, **kwargs):
        # skip when not TypeDecl => PtrDecl is ignored
        if hasattr(node, 'type') and isinstance(node.type, pr.TypeDecl):
            self.handler(node, *args, **kwargs)
        self._recurse_attr(node, 'init', *args, **kwargs)

    def DoWhile(self, node: pr.DoWhile, *args, **kwargs):
        self._recurse_attr(node, 'cond', *args, **kwargs)
        self._recurse_attr(node, 'stmt', *args, **kwargs)

    def For(self, node: pr.For, *args, **kwargs):
        # generally skip control block => use loop_guard
        self._recurse_attr(node, 'stmt', *args, **kwargs)

    def FuncDef(self, node: pr.FuncDef, *args, **kwargs):
        self._recurse_attr(node.decl.type, 'args', *args, **kwargs)
        self._recurse_attr(node, 'body', *args, **kwargs)

    def ID(self, node: pr.ID, *args, **kwargs):
        if node.name not in Variables.RESERVED:
            self.handler(node, *args, **kwargs)

    def If(self, node: pr.If, *args, **kwargs):
        self._recurse_attr(node, 'iftrue', *args, **kwargs)
        self._recurse_attr(node, 'iffalse', *args, **kwargs)

    def Return(self, node: pr.Return, *args, **kwargs):
        # Search return because could be a single-line expr
        self._recurse_attr(node, 'expr', *args, **kwargs)

    def UnaryOp(self, node: pr.UnaryOp, *args, **kwargs):
        if node.op in self.U_OPS:
            self._recurse_attr(node, 'expr', *args, **kwargs)

    def While(self, node: pr.While, *args, **kwargs):
        self._recurse_attr(node, 'cond', *args, **kwargs)
        self._recurse_attr(node, 'stmt', *args, **kwargs)



# ------------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Tuple, Union

from . import Monomial, NODE


class DeltaGraph:
    """
    Delta Graph is a dictionary representing a weighted graph of tuples of
    deltas (also referenced here as a _monomial_list_, a list of Monomials
    without its scalar). We will often refer to tuple of deltas as simple
    node, but a node with length!

    Nodes are "sorted" by this length in order to be compared by chunks of
    same size.

    Weight of edge represents the index where the nodes differ.

    We use tuple because we want them to be hashable (as key in dictionary).

    Example:
        ```
                              ↓
        n1 = ((0,1), (0,2), (0,3), (0,4))
        n2 = ((0,1), (0,2), (1,3), (0,4))
        ```

        in our graph will have:

        ```
        n1 <---- 3 ----> n2
        ```

        or

        ```python
        size = 4   ↓
        graph_dict[4][n1][n2] = 3
        ```

        The graph is symmetric:

        ```python
        graph_dict[4][n2][n1] = 3
        ```

    This representation will help us simplify the evaluation by
    removing redundant/irrelevant choices/paths.

    Attributes:
        graph_dict (dict): Dictionary of nodes.
        degree (int): Graph degree.
    """

    def __init__(
            self,
            *init_nodes: Optional[Union[Monomial, NODE]],
            degree: int = 3
    ):
        """Create a Delta Graph.

        Arguments:
            *init_nodes: Initial list of monomials or nodes.
            degree: Degree of a full node.

        Example:
            Create an empty delta graph

            ```python
            dg = DeltaGraph()
            ```

            Create delta graph with some initial nodes from monomials.

            ```python
            dg = DeltaGraph(mono1, mono2)
            ```
        """
        self.degree = degree
        self.graph_dict = {}
        if init_nodes:
            for node in init_nodes:
                if isinstance(node, Monomial):
                    self.from_monomial(node)
                if isinstance(node, Tuple):
                    self.insert_node(node)

    def __str__(self):
        return "".join(
            [f'{n} {self.graph_dict[n]}\n{26 * "--"}\n' for n in
             self.graph_dict])

    def from_monomial(self, monomial: Monomial) -> None:
        """Add monomial's deltas to the delta graph.

        Arguments:
            monomial: monomial
        """
        self.insert_node(tuple(monomial.deltas))

    def insert_edge(self, node1: NODE, node2: NODE, label: int) -> None:
        """Add an edge of label `label` between `node1` and `node2`
        If one node does not exist, it's created

        Symmetry is also added in the graph.

        Arguments:
            node1: first node
            node2: second node
            label: label over the edge

        """
        size = len(node1)
        if node1 not in self.graph_dict[size]:
            self.graph_dict[size][node1] = {}

        self.graph_dict[size][node1][node2] = label

        # node2 should always exist if add_edge is always called
        # from insert_tuple
        if node2 not in self.graph_dict[size]:
            self.graph_dict[size][node2] = {}
        self.graph_dict[size][node2][node1] = label

    def insert_node(self, node: NODE) -> None:
        """Insert a node into the graph.

        If a node is already in the graph do nothing.
        Else compare it with all nodes of same size with `node_diff`.

        Arguments:
            node: tuple to insert in the graph
        """
        size = len(node)
        # Is it possible that we've already computed that diff in
        # the other way ? (symmetry)
        # Answer: AFA it's called now, NO
        if size not in self.graph_dict:
            self.graph_dict[size] = {}
            self.graph_dict[size][node] = {}
        else:
            # Case may exist where we need to perform simplification
            #    over monomials -> Add here the simplification
            inserted = False
            if node not in self.graph_dict[size]:
                for node2 in list(self.graph_dict[size].keys()):
                    # Already tested when node2[node2][monomial_list] exists
                    # FIXME is it possible ?
                    # if monomial_list not in node2[node2]:
                    diff, i = self.node_diff(node, node2)
                    if diff:
                        inserted = True
                        self.insert_edge(node, node2, i)
                if not inserted:
                    self.graph_dict[size][node] = {}

    @staticmethod
    def remove_index(node: NODE, index: int) -> NODE:
        """Remove delta with given index.

        Example:
            ```
            remove index 4:
            ((0, 2), (1, 3), (2, 4))  ->  ((0, 2), (1, 3))
            ```

        Arguments:
            node: monomial list
            index: index to remove

        Returns:
            a new tuple without deltas with index `index`
        """
        return tuple(filter(lambda x: x[1] != index, node))

    def remove_node(self, node: NODE, index: int) -> None:
        """Remove given node and neighbors connected with same label.

        Also removes edges/labels connected to the node (they no longer exist).
        """
        size = len(node)

        # Keep track of neighbors before removing node
        neighbors = list(self.graph_dict[size][node])

        # Remove node
        del self.graph_dict[size][node]

        # For each neighbour
        for neighbor in neighbors:
            if neighbor in self.graph_dict[size]:
                label = self.graph_dict[size][neighbor][node]
                # If same label then recursively remove neighbour
                if label == index:
                    self.remove_node(neighbor, index)
                # if not just remove the edge
                else:
                    del self.graph_dict[size][neighbor][node]

    @staticmethod
    def node_diff(
            node1: NODE, node2: NODE, index: Optional[int] = None
    ) -> Tuple[bool, int]:
        """Compares two nodes of the same length.

        The return value is a tuple representing the result of comparison
        (boolean) and an index.

        The result is True, if and only if both lists differ only on one
        element regarding the same index. The second return value is the index
        of the corresponding delta.

        Arguments:
            node1: first monomial list
            node2: second monomial list
            index: index with to check number of diff

        Returns:
            diff: boolean True if the lists differ of one element
            i: the index of the delta which differs
        """
        diff_found = False
        i = 0
        while i < len(node1):
            if node1[i] not in node2:
                i1 = node1[i][1]
                # We've already recorded one so two diff
                if diff_found:
                    return False, i1
                # Case recursive call, we've not found yet
                # But index is defined by arguments
                if index is not None:
                    # If the diff has different index than expected
                    if index != i1:
                        return False, index
                    else:
                        # We've found one but still searching more
                        diff_found = True
                # Found first diff without init of index
                else:
                    # Search in other monomial if it has only one diff
                    # Of same index `i1`
                    (diff, _) = DeltaGraph.node_diff(node2, node1, i1)
                    if diff:
                        index = i1
                        # Continue searching see if there are more diff
                        diff_found = True
                    else:
                        return False, i1
            i += 1
        return diff_found, index

    def is_full(self, node: NODE, size: int, index: int) -> bool:
        """Check for cliques of same label.

        Example:
            ```Python
            n3 = ((0, 1), (2, 2), (0, 3))
            n4 = ((0, 1), (2, 2), (1, 3))
            n5 = ((0, 1), (2, 2), (2, 3))

            node = n4
            size = 3
            index = 3
            degree = 3

            n3 -- 3 -- n4
             ⟍         |
                3       3
                  ⟍    |
                    ⟍  |
                      n5

            return True
            ```

        Arguments:
            node: check for clique around that graph node.
            size: size of nodes or graph "level".
            index: index where to find clique.

        Returns:
            True if there is a clique.
        """
        src = self.graph_dict[size][node]
        adjacent = sum([1 for n2 in src if src[n2] == index])
        return adjacent == (self.degree - 1)

    def fusion(self) -> None:
        """Eliminates cliques of same label in a delta graph.

        Example:
            ```python
            m1 = ((0, 1), (0, 2))
            m2 = ((0, 1), (1, 2))
            m3 = ((0, 1), (2, 2), (0, 3))
            m4 = ((0, 1), (2, 2), (1, 3))
            m5 = ((0, 1), (2, 2), (2, 3))

            Delta graph:

              m1 -- 2 -- m2
              m3 -- 3 -- m4
               ⟍         |
                  3      3
                    ⟍    |
                      ⟍  |
                        m5

            Looks for cliques (size default 3) at each index.
            => Graph will simplify to: ((0,1)).
            ```
        """
        # Start from the longest node to the shortest
        for size in sorted(self.graph_dict, reverse=True):
            for node in list(self.graph_dict[size]):
                # For all indexes in deltas of that node
                for index in (list(zip(*node))[1]):
                    # Check if it's full of same index
                    if node in self.graph_dict[size] and \
                            self.is_full(node, size, index):
                        self.remove_node(node, index)
                        self.insert_node(DeltaGraph.remove_index(node, index))

    @property
    def is_empty(self) -> bool:
        return 0 in self.graph_dict and self.graph_dict[0] == {(): {}}



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# flake8: noqa: W605

from __future__ import annotations

import logging
from typing import Optional, Tuple, List, Dict, Union

from . import Choices, DeltaGraph, Polynomial, MATRIX
from . import matrix as matrix_utils
from .semiring import UNIT_MWP, ZERO_MWP

logger = logging.getLogger(__name__)


class Relation:
    """
    A relation is made of a list of variables and a 2D-matrix:

    - Variables of a relation represent the variables of the input
      program under analysis, for example: $X_0, X_1, X_2$.
    - Matrix holds Polynomials and represents the current state of
      the analysis.

    Attributes:
        variables (List[str]): List of variables.
        matrix (MATRIX): Matrix.
    """
    INFTY_BAR = '‖'

    def __init__(self, variables: Optional[List[str]] = None,
                 matrix: Optional[MATRIX] = None):
        """To construct a relation, provide a list of variables
        and an initial matrix.

        If matrix is not provided, the relation matrix will be initialized to
        zero matrix of size matching the number of variables.

        Example:
            Create a new relation from a list of variables:

            ```python
            r = Relation(['X0', 'X1', 'X2'])
            ```

            Creates relation with 0-matrix with and specified variables:

            ```
            X0  |  0  0  0
            X1  |  0  0  0
            X2  |  0  0  0
            ```

        Arguments:
            variables: Program variables.
            matrix: Relation matrix.
        """
        self.variables = [str(v) for v in (variables or []) if v]
        self.matrix = matrix or matrix_utils \
            .init_matrix(len(self.variables))

    @staticmethod
    def identity(variables: List) -> Relation:
        """Create an identity relation.

        This method allows creating a relation whose
        matrix is an identity matrix.

        This is an alternative way to construct a relation.

        Example:
            Create a new identity relation from a list of variables:

            ```python
            r = Relation.identity(['X0', 'X1', 'X2', 'X3'])
            ```

            Creates relation with identity matrix with and variables:

            ```
            X0  |  m  0  0  0
            X1  |  0  m  0  0
            X2  |  0  0  m  0
            X3  |  0  0  0  m
            ```

        Arguments:
            variables: A list of variables.

        Returns:
             Generated relation of given variables and an identity matrix.
        """
        matrix = matrix_utils.identity_matrix(len(variables))
        return Relation(variables, matrix)

    @property
    def is_empty(self):
        return not self.variables or not self.matrix

    @property
    def matrix_size(self):
        return 0 if not self.variables else len(self.variables)

    def __str__(self):
        return Relation.relation_str(self.variables, self.matrix)

    def __add__(self, other):
        return self.sum(other)

    def __mul__(self, other):
        return self.composition(other)

    @staticmethod
    def relation_str(variables: List[str], matrix: MATRIX):
        """Formatted string of variables and matrix."""
        right_pad = len(max(variables, key=len)) if variables else 0
        return '\n'.join(
            [var.ljust(right_pad) + ' | ' + (' '.join(poly)).strip()
             for var, poly in
             [(var, [str(matrix[i][j]) for j in range(len(matrix))])
              for i, var in enumerate(variables)]])

    def replace_column(self, vector: List, variable: str) -> Relation:
        """Replace identity matrix column by a vector.

        Arguments:
            vector: vector by which a matrix column will be replaced.
            variable: program variable, column replacement
                will occur at the index of this variable.

        Raises:
              ValueError: if variable is not found in this relation.

        Returns:
            new relation after applying the column replacement.
        """
        new_relation = Relation.identity(self.variables)
        if variable in self.variables:
            j = self.variables.index(variable)
            for idx, value in enumerate(vector):
                new_relation.matrix[idx][j] = value
        return new_relation

    def while_correction(self, dg: DeltaGraph) -> None:
        """Replace invalid scalars in a matrix by $\\infty$.

        Following the computation of fixpoint for a while loop node, this
        method checks the resulting matrix and replaces all invalid scalars
        with $\\infty$ (W rule in MWP paper):

        - scalar $p$ anywhere in the matrix becomes $\\infty$
        - scalar $w$ at the diagonal becomes $\\infty$

        Example:
            ```text
               Before:                After:

               | m  o  o  o  o |      | m  o  o  o  o |
               | o  w  o  p  o |      | o  i  o  i  o |
               | o  o  m  o  o |      | o  o  m  o  o |
               | w  o  o  m  o |      | w  o  o  m  o |
               | o  o  o  o  p |      | o  o  o  o  i |
            ```

        Related discussion: [issue #14](
        https://github.com/statycc/pymwp/issues/14).

        Arguments:
            dg: DeltaGraph instance.
        """
        for i, vector in enumerate(self.matrix):
            for j, poly in enumerate(vector):
                for mon in poly.list:
                    if mon.scalar == "p" or (mon.scalar == "w" and i == j):
                        mon.scalar = "i"
                        dg.from_monomial(mon)

    def loop_correction(self, x_var: str, dg: DeltaGraph) -> None:
        """Loop correction to replace invalid scalars by $\\infty$.

        Following computation of a loop fixpoint, this method checks
        the resulting matrix by rule L: scalars >$m$ at the diagonal
        become $\\infty$. If exists M$_ij$ = p then row X, col j => p.

        Related discussion [issue #5](
        https://github.com/statycc/pymwp/issues/5).

        Arguments:
            x_var: Loop control variable.
            dg: DeltaGraph instance.
        """
        ell = self.variables.index(x_var)
        for i, vector in enumerate(self.matrix):
            for j, poly in enumerate(vector):
                for mon in poly.list:
                    if i == j and mon.scalar != "m":
                        mon.scalar = "i"
                        dg.from_monomial(mon)
                    if mon.scalar == "p":
                        self.matrix[ell][j] = self.matrix[ell][j] \
                            .add(Polynomial(mon.copy()))

    def sum(self, other: Relation) -> Relation:
        """Sum two relations.

        Calling this method is equivalent to syntax `relation + relation`.

        Arguments:
            other: Relation to sum with self.

        Returns:
           A new relation that is a sum of inputs.
        """
        er1, er2 = Relation.homogenisation(self, other)
        new_matrix = matrix_utils.matrix_sum(er1.matrix, er2.matrix)
        return Relation(er1.variables, new_matrix)

    def composition(self, other: Relation) -> Relation:
        """Composition of current and another relation.

        Calling this method is equivalent to syntax `relation * relation`.

        Composition will:

        1. combine the variables of two relations, and
        2. produce a single matrix that is the product of matrices of
            the two input relations.

        Arguments:
            other: Relation to compose with current

        Returns:
           a new relation that is a product of inputs.
        """
        logger.debug("starting composition...")
        er1, er2 = Relation.homogenisation(self, other)
        logger.debug("composing matrix product...")
        new_matrix = matrix_utils.matrix_prod(er1.matrix, er2.matrix)
        logger.debug("...relation composition done!")
        return Relation(er1.variables, new_matrix)

    def equal(self, other: Relation) -> bool:
        """Determine if two relations are equal.

        For two relations to be equal they must have:

        1. the same variables (independent of order), and
        2. matrix polynomials must be equal element-wise.

        Arguments:
            other: relation to compare

        Returns:
            true when two relations are equal
            and false otherwise.
        """

        # must have same variables in same order
        if set(self.variables) != set(other.variables):
            return False

        # not sure homogenisation is necessary here
        # --> yes we need it
        er1, er2 = Relation.homogenisation(self, other)

        for row1, row2 in zip(er1.matrix, er2.matrix):
            for poly1, poly2 in zip(row1, row2):
                if poly1 != poly2:
                    return False
        return True

    def fixpoint(self) -> Relation:
        """Compute sum of compositions until no changes occur.

        Returns:
            resulting relation.
        """
        fix_vars = self.variables
        matrix = matrix_utils.identity_matrix(len(fix_vars))
        fix = Relation(fix_vars, matrix)
        prev_fix = Relation(fix_vars, matrix)
        current = Relation(fix_vars, matrix)
        logger.debug(f"computing fixpoint for variables {fix_vars}")

        while True:
            prev_fix.matrix = fix.matrix
            current = current * self
            fix = fix + current
            if fix.equal(prev_fix):
                logger.debug(f"fixpoint done {fix_vars}")
                return fix

    def apply_choice(self, *choices: int) -> SimpleRelation:
        """Get the matrix corresponding to provided sequence of choices.

        Arguments:
            choices: Tuple of choices.

        Returns:
            New relation with simple-values matrix of scalars.
        """
        new_mat = [[self.matrix[i][j].choice_scalar(
            *choices, least_scalar=UNIT_MWP if i == j else ZERO_MWP)
            for j in range(self.matrix_size)]
            for i in range(self.matrix_size)]
        return SimpleRelation(self.variables.copy(), matrix=new_mat)

    def infty_vars(self, only_incl: List[str] = None) \
            -> Dict[str, List[str]]:
        """Identify all variable pairs that for some choices, can raise
        infinity result.

        Returns:
            Dictionary of potentially infinite dependencies, where
                the key is source variable and value is list of targets.
                All entries are non-empty.
        """
        return dict([(x, y) for x, y in [
            (src, [tgt for tgt, p in zip(self.variables, polys)
                   if p.some_infty and
                   (not only_incl or src in only_incl or tgt in only_incl)])
            for src, polys in zip(self.variables, self.matrix)]
                     if len(y) != 0])

    def infty_pairs(self, only_incl: List[str] = None) -> str:
        """List of potential infinity dependencies."""
        fmt = [f'{s} ➔ {", ".join(t)}'
               for s, t in self.infty_vars(only_incl).items()]
        return f' {Relation.INFTY_BAR} '.join(fmt)

    def to_dict(self) -> dict:
        """Get dictionary representation of a relation."""
        return {"matrix": matrix_utils.encode(self.matrix)}

    def show(self):
        """Display relation."""
        print(str(self))

    @staticmethod
    def homogenisation(r1: Relation, r2: Relation) \
            -> Tuple[Relation, Relation]:
        """Performs homogenisation on two relations.

        After this operation both relations will have same
        variables and their matrices of the same size.

        This operation will internally resize matrices as needed.

        Arguments:
            r1: First relation to homogenise.
            r2: Second relation to homogenise.

        Returns:
            Homogenised versions of the 2 inputs relations.
        """

        # check equality
        if r1.variables == r2.variables:
            return r1, r2

        # check empty cases
        if r1.is_empty:
            return Relation.identity(r2.variables), r2

        if r2.is_empty:
            return r1, Relation.identity(r1.variables)

        logger.debug("matrix homogenisation...")

        # build a list of all distinct variables; maintain order
        extended_vars = r1.variables + [v for v in r2.variables
                                        if v not in r1.variables]

        # resize matrices to match new number of variables
        new_matrix_size = len(extended_vars)

        # first matrix: just resize -> this one is now done
        matrix1 = matrix_utils.resize(r1.matrix, new_matrix_size)

        # second matrix: create and initialize as identity matrix
        matrix2 = matrix_utils.identity_matrix(new_matrix_size)

        # index of each extended_vars iff variable exists in r2.
        # we will use this to mapping from old -> new matrix to fill
        # the new matrix; the indices may be in different order.
        index_dict = {index: r2.variables.index(var)
                      for index, var in enumerate(extended_vars)
                      if var in r2.variables}

        # generate a list of all valid <row, column> combinations
        index_map = [t1 + t2 for t1 in index_dict.items()
                     for t2 in index_dict.items()]

        # fill the resized matrix with values from original matrix
        for mj, rj, mi, ri in index_map:
            matrix2[mi][mj] = r2.matrix[ri][rj]

        return Relation(extended_vars, matrix1), Relation(extended_vars,
                                                          matrix2)

    def eval(self, choices: List[int], index: int, *scalars: str) -> Choices:
        """Evaluate program matrix for possible derivation choices.

        Arguments:
            choices: List of choices at each index, `[0,1,2]`.
            index: Accumulated program counter.
            scalars: Exclude specified scalars.

        Returns:
            A choice object for the evaluated matrix.
        """
        infinity_deltas = set()

        # get all choices leading to infinity
        for row in self.matrix:
            for poly in row:
                infinity_deltas.update(poly.eval(*scalars))

        # generate valid choices
        return Choices.generate(choices, index, infinity_deltas)

    def var_eval(self, choices: List[int], index: int,
                 variables: Union[str, List[str]] = None,
                 *scalars: str) -> Union[Choices, Dict[str, Choices]]:
        """Evaluate choices for each individual variable.

        This is same as `eval`, except it generates the choice-vectors
        column-wise.

        Arguments:
            choices: List of choices at each index, `[0,1,2]`
            index: Accumulated program counter.
            variables: One or more variables to evaluate.
            scalars: Exclude specified scalars.

        Returns:
            A dictionary where the key is a variable name,
            and value is a choice object for the evaluated variable.
        """
        result = {}
        one_var = isinstance(variables, str) and len(variables)
        eval_set = [variables] if one_var else (variables or self.variables)
        for v_name in eval_set:
            d, col = set(), self.variables.index(v_name)
            for row in self.matrix:
                d.update(row[col].eval(*scalars))
            result[v_name] = Choices.generate(choices, index, d)
        return result[variables] if one_var else result


class SimpleRelation(Relation):
    """Specialized instance of relation, where matrix contains only
       scalar values, no polynomials.

    A relation converts to a SimpleRelation by applying a
    derivation choice, see: `Relation.apply_choice`.
    """

    def __init__(self, variables: Optional[List[str]] = None,
                 matrix: Optional[List[List[str]]] = None):
        super().__init__(variables, matrix)



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# flake8: noqa: W605

import logging
from functools import reduce
from typing import Any, Optional, List

from . import Monomial, Polynomial, MATRIX
from .semiring import ZERO_MWP, UNIT_MWP

ZERO = Polynomial(ZERO_MWP)

UNIT = Polynomial(UNIT_MWP)

logger = logging.getLogger(__name__)


def init_matrix(size: int, init_value: Optional[Any] = None) -> List[list]:
    """Create empty matrix of specified size.

    Example:
        Generate 5 x 5 size zero-matrix.

        ```python
        init_matrix(5)
        ```

        Generates:

        ```python
        [[o, o, o, o, o],
         [o, o, o, o, o],
         [o, o, o, o, o],
         [o, o, o, o, o],
         [o, o, o, o, o]]
        ```

    Arguments:
        size: matrix size.
        init_value: value to place at each index. If not provided,
            will default to 0-polynomial.

    Returns:
        Initialized matrix.
    """
    value = init_value if init_value is not None else ZERO
    return [[value for _ in range(size)] for _ in range(size)]


def identity_matrix(size: int) -> List[list]:
    """Create identity matrix of specified size.

    Example:
        Generate 5 x 5 size identity matrix:

        ```python
        identity_matrix(5)
        ```

        Generates:

        ```python
        [[m, o, o, o, o],
         [o, m, o, o, o],
         [o, o, m, o, o],
         [o, o, o, m, o],
         [o, o, o, o, m]]
        ```

    Arguments:
        size: matrix size

    Returns:
        New identity matrix.
    """
    return [[UNIT if i == j else ZERO
             for j in range(size)] for i in range(size)]


def encode(matrix: MATRIX) -> List[List[List[dict]]]:
    """Converts a matrix of polynomials to a matrix of dictionaries.

    This function is useful when preparing to write a matrix of polynomials to
    a file. The same matrix can later be restored using matrix
    [decode](matrix.md#pymwp.matrix.decode).

    Arguments:
        matrix: matrix to encode

    Raises:
        AttributeError: If the matrix does not contain polynomials.

    Returns:
        Encoded matrix.
    """
    return [[
        [mono.to_dict() for mono in polynomial.list]
        for polynomial in row]
        for (i, row) in enumerate(matrix)]


def decode(matrix: List[List[List[dict]]]) -> MATRIX:
    """Converts matrix of dictionaries to a matrix of polynomials.

    Primary use case of this function is for restoring a matrix of
     polynomials from a file (assuming [encode](matrix.md#pymwp.matrix.encode)
     was used to generate that file).

    Arguments:
        matrix: matrix to decode

    Raises:
        TypeError: If the matrix value is not iterable
        AttributeError: If the matrix elements are not
            valid encoded polynomials.

    Returns:
        Decoded matrix of polynomials.
    """
    return [[
        Polynomial(*[Monomial(
            scalar=monomial["scalar"],
            deltas=monomial["deltas"])
            for monomial in polynomial])
        for polynomial in row]
        for (i, row) in enumerate(matrix)]


def matrix_sum(matrix1: MATRIX, matrix2: MATRIX) -> MATRIX:
    """Compute the sum of two polynomial matrices.

    Arguments:
        matrix1: First polynomial matrix.
        matrix2: Second polynomial matrix.

    Returns:
        A new matrix that represents the sum of the two inputs.
    """

    return [[matrix1[i][j] + matrix2[i][j]
             for j in range(len(matrix1))]
            for i in range(len(matrix1))]


def matrix_prod(matrix1: MATRIX, matrix2: MATRIX) -> MATRIX:
    """Compute the product of two polynomial matrices.

    Arguments:
        matrix1: First polynomial matrix.
        matrix2: Second polynomial matrix.

    Returns:
        A new matrix that represents the product of the two inputs.
    """

    return [[
        reduce(lambda total, k:
               total + (matrix1[i][k] * matrix2[k][j]),
               range(len(matrix1)), ZERO)
        for j in range(len(matrix2))]
        for i in range(len(matrix1))]


def resize(matrix: MATRIX, new_size: int) -> MATRIX:
    """Create a new matrix of polynomials of specified size.

    The resized matrix is initialized as an identity matrix
    then filled with values from the original matrix.

    Arguments:
        matrix: original matrix
        new_size: width/height of new matrix

    Returns:
        New matrix of specified size, filled with values from
            the original matrix.
    """
    res = identity_matrix(new_size)
    bound = min(new_size, len(matrix))
    for i in range(bound):
        for j in range(bound):
            res[i][j] = matrix[i][j]
    return res


def show(matrix: MATRIX, prefix: str = None, postfix: str = None) -> None:
    """Pretty print a matrix at the screen.

    Using the keyword arguments to display additional text before or
    after the matrix.

    Example:
        === "Matrix only"

            ```python
            my_matrix = identity_matrix(3)
            show(my_matrix)
            ```

            Displays:

            ```
            ['  +m', '  +o', '  +o']
            ['  +o', '  +m', '  +o']
            ['  +o', '  +o', '  +m']
            ```

        === "Matrix with text"

            ```python
            my_matrix = identity_matrix(3)
            header = '|   x1   |   x2  |  x3 |'
            show(my_matrix, prefix=header)
            ```

            Displays:

            ```
            |   x1   |   x2  |  x3 |
            ['  +m', '  +o', '  +o']
            ['  +o', '  +m', '  +o']
            ['  +o', '  +o', '  +m']
            ```

    Arguments:
        matrix: The matrix to display.
        prefix: display some text before displaying matrix
        postfix: display some text after displaying matrix

    Raises:
        TypeError: If the matrix is not iterable (type list of lists)
    """
    if prefix:
        print(prefix)
    for row in matrix:
        print([str(r) for r in row])
    if postfix:
        print(postfix)
    print(' ')


def equals(matrix1: MATRIX, matrix2: MATRIX) -> bool:
    """Determine if two matrices are equal.

    This function performs element-wise equality comparisons on values of
    two matrices. The two matrices must be the same size. For any two matrices
    of different size the result is always `False`.

    This function can evaluate values that are comparable by equals `==`
    operator.

    Arguments:
        matrix1: First matrix.
        matrix2: Second matrix.

    Raises:
        TypeError: If the matrix value is not iterable.

    Returns:
        `True` if matrices are equal element-wise and `False` otherwise.
    """
    # equal size
    if [len(row) for row in matrix1] != [len(row) for row in matrix2]:
        return False

    # element-wise comparison
    for row_index, column in enumerate(matrix1):
        for col_index, value in enumerate(column):
            if matrix2[row_index][col_index] != value:
                return False
    return True


def fixpoint(matrix: MATRIX) -> MATRIX:
    """Computes the star operation $1 + M + M^2 + M^3 + …$

    This function assumes provided input is a square matrix.

    Arguments:
        matrix: Matrix for which to compute fixpoint.

    Returns:
        $M^*$
    """
    _1_ = identity_matrix(len(matrix))
    previous = matrix
    next_matrix = matrix
    result = matrix_sum(_1_, matrix)

    while not equals(previous, result):
        previous = result
        next_matrix = matrix_prod(next_matrix, matrix)  # M^2, M^3, M^4....
        result = matrix_sum(result, next_matrix)

    return result



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Dict, Union, List, Tuple, Any, Callable, Type

from pymwp import Relation, Bound, MwpBound, Choices
from .matrix import decode

logger = logging.getLogger(__name__)


class Timeable:
    """Represents an entity whose runtime can be measured.

    Attributes:
        start_time (int): recorded start time.
        end_time (int): recorded end time.
    """

    def __init__(self):
        self.start_time = 0
        self.end_time = 0

    @property
    def time_diff(self) -> int:
        """Time delta between analysis start and end time."""
        return self.end_time - self.start_time

    @property
    def dur_s(self) -> float:
        """Duration in seconds."""
        return round(self.time_diff / 1e9, 1)

    @property
    def dur_ms(self) -> int:
        """Duration in milliseconds."""
        return int(self.time_diff / 1e6)

    def on_start(self) -> Timeable:
        """Called at start of timeable entity."""
        self.start_time = time.time_ns()
        return self

    def on_end(self) -> Timeable:
        """Called at end of timeable entity."""
        self.end_time = time.time_ns()
        return self


class Serializable(ABC):
    """General utilities for converting results to JSON-writable
    objects and vice versa."""

    @property
    def _attrs(self) -> List[str]:
        """List of simple attribute names."""
        return []

    @property
    def _ser_attrs(self) -> List[Tuple[str, Type[Serializable]]]:
        """Attributes of type Serializable."""
        return []

    @property
    def _ser_dict(self) -> List[Tuple[str, str, Type[Serializable]]]:
        """Dictionaries attributes where values are of Serializable type."""
        return []

    @property
    def _ser_list(self) -> List[Tuple[str, Type[Serializable]]]:
        """List attributes where values are of Serializable type."""
        return []

    def to_dict(self) -> Dict[str, Any]:
        """Convert an object to a JSON-compatible dictionary."""
        # simple type attributes
        sv = map(self.__getattribute__, self._attrs)
        simple = dict(zip(self._attrs, sv))

        # serializable types
        ok = [k for k, _ in self._ser_attrs]
        ov = [self.__getattribute__(k).to_dict() for k in ok]
        objs = dict(zip(ok, ov))

        # lists of serializable types
        lk = [k for k, _ in self._ser_list if getattr(self, k)]
        lv = [[v.to_dict() for v in getattr(self, k)] for k in lk]
        lists = dict(zip(lk, lv))

        # dictionary types
        dk = [n for n, _, _ in self._ser_dict if getattr(self, n)]
        dicts = dict(zip(dk, [dict([
            (n, v.to_dict()) for n, v in getattr(self, k).items()])
            for k in dk]))

        return {**simple, **objs, **lists, **dicts}

    @staticmethod
    @abstractmethod
    def from_dict(**kwargs) -> Serializable:
        """Restore object from a dictionary; reverses `to_dict()`."""
        pass

    @staticmethod
    def _load(obj: Serializable, **kwargs) -> Any:
        """Initializes a Serializable object from kwargs.

        Arguments:
            obj: Initialized object.
            *kwargs: Object data.

        Returns:
            Loaded object.
        """
        for key in obj._attrs:
            obj._try_set(obj, key, **kwargs)
        for attr, objT in obj._ser_attrs:
            values = Serializable._try_get(attr, **kwargs)
            setattr(obj, attr, objT.from_dict(**values) if values else None)
        for attr, objT in obj._ser_list:
            values = Serializable._try_get(attr, **kwargs) or []
            setattr(obj, attr, [objT.from_dict(**v) for v in values])
        for attr, key, objT in obj._ser_dict:
            items = Result._try_get(attr, **kwargs) or {}
            values = [objT.from_dict(**value) for value in items.values()]
            keys = [getattr(obj, key) for obj in values]
            setattr(obj, attr, dict(zip(keys, values)))
        return obj

    @staticmethod
    def _try_set(target: object, attr: str, *keys: str, **kwargs) -> None:
        """Try set target.attr from kwargs matching keys."""
        keys_ = (attr,) if not keys else keys
        ob = Serializable._try_get(*keys_, **kwargs)
        setattr(target, attr, ob) if ob else None

    @staticmethod
    def _try_get(*keys: str, **kwargs) -> Any:
        """Try to get a kwargs value that matches keys."""
        ob = kwargs
        for i, key in enumerate(keys):
            ob = ob[key] if (ob and key in ob) else None
        return ob


class Program(Serializable):
    """Details about analyzed C-language input file.

    Attributes:
        n_lines (int): Lines of code in input program.
        program_path (str): Path to program file.
    """

    def __init__(self, n_lines: int = -1, program_path: str = None):
        self.n_lines: int = n_lines
        self.program_path: str = program_path

    @property
    def _attrs(self) -> List[str]:
        """List of attributes."""
        return ['n_lines', 'program_path']

    @property
    def name(self) -> Optional[str]:
        """Get program name without path and extension."""
        return Path(self.program_path).stem if self.program_path else None

    @staticmethod
    def from_dict(**kwargs) -> Program:
        """Restore Program object."""
        return Serializable._load(Program(), **kwargs)


class FuncResult(Timeable, Serializable):
    """Analysis results for one function of the input program.

    Attributes:
        name (str): Function name.
        infinite (bool): True if no valid derivation exists.
        variables (List[str]): List of program variables.
        relation (Relation): Relation object; does not
        choices (Choices): A choice vector-object.
        bound (Bound): A bound of mwp-bounds.
        inf_flows (str): Description of problematic flows.
        index (int): Degree of derivation choice.
        func_code (str): Function source code.
    """

    def __init__(self, name: str, infinite: bool = False,
                 variables: Optional[List[str]] = None,
                 relation: Optional[Relation] = None,
                 choices: Optional[Choices] = None,
                 bound: Optional[Bound] = None,
                 inf_flows: Optional[str] = None,
                 index: int = -1, func_code: str = None):
        super().__init__()
        self.name: str = name
        self.infinite: bool = infinite
        self.variables: List[str] = variables or []
        self.relation: Relation = relation
        self.choices: Choices = choices
        self.bound: Bound = bound
        self.inf_flows: str = inf_flows
        self.index: int = index
        self.func_code: str = func_code

    def __str__(self):
        if not self.infinite and not self.bound:
            return 'Some bound exists'
        txt = f'function: {self.name}'
        txt += f' • time: {self.dur_ms:,} ms\n'
        txt += f'variables: {len(self.variables)}'
        if self.infinite:
            txt += ' • num-bounds: '
            txt += '0 (infinite)'
            if self.inf_flows:
                txt += f'\nProblematic flows: {self.inf_flows}'
        elif len(self.variables):
            txt += ' • num-bounds: '
            txt += f'{self.n_bounds:,}\n'
            txt += f'{Bound.show(self.bound, True, True)}'
        return txt

    @property
    def _attrs(self) -> List[str]:
        """List of attribute names."""
        return 'name,infinite,start_time,end_time,variables,' \
               'inf_flows,index,func_code'.split(',')

    @property
    def n_vars(self) -> int:
        """Number of variables."""
        return len(self.variables)

    @property
    def n_bounds(self) -> int:
        """Number of bounds."""
        return self.choices.n_bounds if self.choices else 0

    def to_dict(self) -> dict:
        """Serialize a function result."""
        result = super().to_dict()
        if self.relation:
            result['relation'] = self.relation.to_dict()
        if self.choices:
            result['choices'] = self.choices.valid
        if self.bound:
            result['bound'] = self.bound.to_dict()
        return result

    @staticmethod
    def from_dict(name: str = None, **kwargs) -> FuncResult:
        """Deserialize a function result."""
        func = Serializable._load(FuncResult(name), **kwargs)
        matrix = FuncResult._try_get('relation', 'matrix', **kwargs)
        if matrix:
            func.relation = Relation(func.variables, decode(matrix))
        choices = FuncResult._try_get('choices', **kwargs)
        if choices:
            func.choices = Choices(choices)
        bound = FuncResult._try_get('bound', **kwargs)
        if bound:
            func.bound = Bound(bound)
        return func


class FuncLoops(Timeable, Serializable):
    """Analysis result for a function with loops, when running
    loop analysis mode. `FuncLoops` captures the results for all loops
    inside their parent function.

    Attributes:
        name (str): Containing function name.
        loops (Dict[str, LoopResult]): Function loop analysis results.
    """

    def __init__(self, name: str = None):
        super().__init__()
        self.name: str = name
        self.loops: List[LoopResult] = []

    def __str__(self):
        loops = [f'{i}. {lp}' for i, lp in enumerate(self.loops)]
        lp_str = ('\n# ' + '\n# '.join(loops)) if loops else ''
        return (f'function: {self.name} • loops: {self.n_loops}'
                f' • time: {self.dur_ms:,} ms{lp_str}')

    @property
    def _attrs(self) -> List[str]:
        return ['name', 'start_time', 'end_time']

    @property
    def _ser_list(self) -> List[Tuple[str, Type[Serializable]]]:
        return [('loops', LoopResult)]

    @property
    def n_loops(self) -> int:
        """Number of analyzed loops"""
        return len(self.loops)

    @staticmethod
    def from_dict(**kwargs) -> FuncLoops:
        """Restore FuncLoops object."""
        return Serializable._load(FuncLoops(), **kwargs)


class LoopResult(Timeable, Serializable):
    """Analysis result for one loop.

    Attributes:
        loop_code (str): The analyzed loop.
        variables (Dict[str, VResult]): Results by variable.
    """

    def __init__(self, loop_code: str = None):
        super().__init__()
        self.loop_code: str = loop_code
        self.variables: Dict[str, VResult] = {}

    def __str__(self):
        txt = f'{self.loop_desc}'
        if not self.variables:
            txt += f'\nvariables: {len(self.variables)}'
        else:
            txt += f'\nlinear: {", ".join(self.linear) or "—"}'
            txt += f'\nindependent: {", ".join(self.weak) or "—"}'
            txt += f'\npolynomial: {", ".join(self.poly) or "—"}'
            txt += f'\ninfinity: {", ".join(self.exp) or "—"}'
            txt += f'\n{self.as_bound.show(True, True)}'
        return txt

    @property
    def _attrs(self) -> List[str]:
        return ['loop_code', 'start_time', 'end_time']

    @property
    def _ser_dict(self) -> List[Tuple[str, str, Type[Serializable]]]:
        return [('variables', 'name', VResult)]

    @property
    def n_vars(self) -> int:
        """Number of variables."""
        return len(self.variables)

    @property
    def n_bounded(self) -> int:
        """Number of variables with a known bound."""
        exp = filter(lambda x: x.exponential, self.variables.values())
        return self.n_vars - len(list(exp))

    @property
    def n_lines(self) -> int:
        """Number of code lines in a loop."""
        return self.loop_code.strip().count('\n')

    @property
    def loop_desc(self) -> str:
        """Loop description => header block."""
        header = self.loop_code.split('\n', 1)[0].strip()
        return (header[:40] if len(header) > 40 else header) + '…'

    @property
    def linear(self) -> List[str]:
        """All variables with a linear bound."""
        return self.var_sat(lambda r: r.is_m)

    @property
    def weak(self) -> List[str]:
        """All variables with a weak polynomial bound."""
        return self.var_sat(lambda r: r.is_w and not r.is_m)

    @property
    def poly(self) -> List[str]:
        """All variables with a polynomial bound."""
        return self.var_sat(lambda r: r.is_p and not r.is_w)

    @property
    def exp(self) -> List[str]:
        """All variables with an unknown bound."""
        return self.var_sat(lambda r: r.exponential)

    @property
    def as_bound(self) -> Bound:
        """Combines variable results to a Bound-type."""
        return Bound(dict([
            (v, res.bound.bound_str) for v, res in
            self.variables.items() if res.bound]))

    def var_sat(self, cond: Callable[[VResult], bool]) -> List[str]:
        """List of variables satisfying a condition."""
        return [v for v, r in self.variables.items() if cond(r)]

    @staticmethod
    def from_dict(**kwargs) -> LoopResult:
        """Restore LoopResult object."""
        return Serializable._load(LoopResult(), **kwargs)


class VResult(Serializable):
    """Analysis result for a single variable.

    Attributes:
        name (str): Variable name.
        is_m (bool): Has maximal linear bound.
        is_w (bool): Has weak polynomial bound.
        is_p (bool): Has polynomial bound.
        bound (Optional[MwpBound]): A bound (if exists).
        choices (Optional[Choice]): Bound choices.
    """

    def __init__(self, name: str = None, is_m: bool = False,
                 is_w: bool = False, is_p: bool = False,
                 bound: MwpBound = None, choices: Choices = None):
        self.name = name
        self.bound = bound
        self.choices = choices
        self._is_m = False
        self._is_w = False
        self._is_p = False
        self.is_m = is_m
        self.is_w = is_w
        self.is_p = is_p

    @property
    def _attrs(self) -> List[str]:
        return 'name,is_m,is_w,is_p'.split(',')

    @property
    def exponential(self):
        """Variable has no known bound (polynomial or less)."""
        return self.is_p is False

    @property
    def is_m(self) -> bool:
        """Variable has a max of linear bound."""
        return self._is_m

    @is_m.setter
    def is_m(self, value: bool):
        if value:
            self._is_m = self._is_w = self._is_p = True
        else:
            self._is_m = False

    @property
    def is_w(self) -> bool:
        """Variable has a weak polynomial bound."""
        return self._is_w

    @is_w.setter
    def is_w(self, value: bool):
        if value:
            self._is_w = self._is_p = True
        else:
            self._is_m = self._is_w = False

    @property
    def is_p(self) -> bool:
        """Variable has a polynomial bound."""
        return self._is_p

    @is_p.setter
    def is_p(self, value: bool):
        if value:
            self._is_p = True
        else:
            self._is_m = self._is_w = self._is_p = False

    def to_dict(self) -> dict:
        """Serialize a VResult object."""
        result = super().to_dict()
        if self.choices:
            result['choices'] = self.choices.valid
        if self.bound:
            result['bound'] = self.bound.bound_str
        return result

    @staticmethod
    def from_dict(**kwargs) -> VResult:
        """Restore VResult object."""
        r = Serializable._load(VResult(), **kwargs)
        choices = VResult._try_get('choices', **kwargs)
        r.choices = Choices(choices) if choices else None
        bound = VResult._try_get('bound', **kwargs)
        r.bound = MwpBound(bound) if bound else None
        return r


class Result(Timeable, Serializable):
    """Captures analysis result and details about the analysis process.

    Attributes:
        program (Program): Information about analyzed C File.
        relations (Dict[str, FuncResult]): Dictionary of function results.
        loops (Dict[str, FuncLoops]): Dictionary of function loop results.
    """

    def __init__(self):
        super().__init__()
        self.program: Program = Program()
        self.relations: Dict[str, FuncResult] = {}
        self.loops: Dict[str, FuncLoops] = {}

    @property
    def _attrs(self) -> List[str]:
        """List of attribute names."""
        return ['start_time', 'end_time']

    @property
    def _ser_attrs(self) -> List[Tuple[str, Type[Serializable]]]:
        return [('program', Program)]

    @property
    def _ser_dict(self) -> List[Tuple[str, str, Type[Serializable]]]:
        return [('loops', 'name', FuncLoops),
                ('relations', 'name', FuncResult)]

    @property
    def n_functions(self) -> int:
        """Number of functions in analyzed program."""
        return len(self.relations.keys())

    @property
    def n_loops(self) -> int:
        """Number of loops in functions of analyzed program."""
        return sum([lp.n_loops for lp in self.loops.values()])

    def add_relation(self, result: FuncResult) -> None:
        """Appends function analysis to result."""
        self.relations[result.name] = result
        Result.pretty_print(str(result))

    def add_loop(self, result: FuncLoops) -> None:
        """Append loop analysis to result."""
        self.loops[result.name] = result
        Result.pretty_print(str(result))

    @staticmethod
    def pretty_print(txt: str, line_w: int = 50, hb: str = '─') -> str:
        """Draws a colored box around text before display.

        Arguments:
            txt: Some text to display.
            line_w: Formatted text line width.
            hb: Horizontal bar box-drawing character

        Returns:
             Formatted text.
        """
        color, endc = '\033[96m', '\033[0m'
        top_bar = (hb * (line_w + 3))
        bot_bar = top_bar[:]
        land, i_bar = Bound.LAND, Relation.INFTY_BAR
        lines, fst_land = [], True
        for vals in txt.split('\n'):
            while vals:
                # don't wrap if bound expr fits in one line
                fits = fst_land and len(vals) < line_w
                # find ideal line break index
                if land in vals[:line_w] and not fits:
                    split_at = 1 + vals[:line_w].index(land)
                elif i_bar in vals[:line_w] and len(vals) > line_w:
                    split_at = 1 + vals[:line_w].rindex(i_bar)
                elif ' ' in vals[:line_w] and len(vals) > line_w:
                    split_at = 1 + vals[:line_w].rindex(' ')
                else:
                    split_at = line_w
                # split to current...remaining
                part = vals[:split_at].strip()
                vals = vals[split_at:].strip()
                fst_land = fst_land and (
                        land not in part and i_bar not in part)
                # format line and append
                lines += [f' {part:<{line_w}}']
        parts = '\n'.join([top_bar, '\n'.join(lines), bot_bar])
        logger.info(f'\n{color}{parts}{endc}')
        return parts

    def get_func(self, name: Optional[str] = None) \
            -> Union[FuncResult, FuncLoops, Dict[str, FuncResult],
                     Dict[str, FuncLoops]]:
        """Returns analysis result for function(s).

        Here "analysis" means either whole-function analysis, or loop
        analysis, based on executed analysis mode; they cannot co-exist in
        the same result.

        * If `name` argument is provided and key exists, returns a result
         for exact value match.
        * If program contained exactly 1 function, returns result for that
          function.
        * Otherwise, returns a dictionary of results for each analyzed
          function, as in: `<function_name, analysis_result>`

        Arguments:
            name: Name of function.

        Returns:
            A function analysis result, or a dictionary of results.
        """
        if name:
            if name in self.relations:
                return self.relations[name]
            if name in self.loops:
                return self.loops[name]
        if self.n_functions == 1:
            key = next(iter(self.relations))
            return self.relations[key]
        if self.n_loops == 1:
            key = next(iter(self.loops))
            return self.loops[key]
        return self.relations if self.relations else self.loops

    def log_result(self) -> Result:
        """Display here all interesting stats about analysis result."""
        if self.n_functions == 0 and self.n_loops == 0:
            logger.warning("Nothing was analyzed")
        logger.info(f'Total time: {self.dur_s} s ({self.dur_ms} ms)')
        return self

    @staticmethod
    def from_dict(**kwargs) -> Result:
        """Restore Result object."""
        return Serializable._load(Result(), **kwargs)



# ------------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from enum import IntEnum
from typing import Tuple, List, TypeVar

DELTA = Tuple[int, int]
"""Delta is a tuple of two integers: (choice, index)."""

DELTAS = TypeVar('DELTAS', bound=DELTA)
"""A TypeVar for deltas."""

SEQ = NODE = Tuple[DELTA, ...]
"""A tuple of deltas."""

VECT = Tuple[Tuple[int, ...], ...]
"""A vector is a tuple of integer-tuples."""

CHOICES = List[List[List[int]]]
"""A list of choice vectors."""

# flake8: noqa: F821
COM_RES = Tuple[int, 'RelationList', bool]
"""Command analysis result type."""

B_TRIPLE = Tuple[Tuple[str, ...], Tuple[str, ...], Tuple[str, ...]]
"""mwp-bound triple with three variable lists."""

# flake8: noqa: F821
MATRIX = List[List['Polynomial']]
"""mwp-matrix type."""


class Comparison(IntEnum):
    """Represent result of delta comparison."""
    SMALLER = 0
    EQUAL = 1
    LARGER = 2


class SetInclusion(IntEnum):
    """Represent result of monomial inclusion."""
    EMPTY = 0
    CONTAINS = 1
    INCLUDED = -1



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import json
import logging
import os

from . import Result

logger = logging.getLogger(__name__)


def loc(input_file: str) -> int:
    """Get number of lines is a file.

    Arguments:
        input_file: Path to input file.

    Returns:
        Number of non-empty lines in input file.
    """
    with open(input_file, 'r') as fp:
        lines = fp.readlines()
    lines = [1 for line in lines if line and len(line.strip())]
    return sum(lines)


def default_file_out(input_file: str) -> str:
    """Generates default output file.

    Arguments:
        input_file: input filename (with or without path).

    Returns:
        Generated output filename with path.
    """
    file_only = os.path.splitext(input_file)[0]
    file_name = os.path.basename(file_only)
    return os.path.join("output", f"{file_name}.json")


def save_result(file_name: str, analysis_result: Result) -> None:
    """Save analysis result to file as JSON.

    Expected behavior:

    - if path to output file does not exist it will be created
    - if output file does not exist it will be created
    - if output file exists it will be overwritten

    Arguments:
        file_name: Filename where to write.
        analysis_result: Result object.
    """
    # JSON serializable object
    file_content = analysis_result.to_dict()

    # ensure directory path exists
    dir_path, _ = os.path.split(file_name)
    if len(dir_path) > 0 and not os.path.exists(dir_path):
        os.makedirs(dir_path)

    # write to file
    with open(file_name, "w") as outfile:
        json.dump(file_content, outfile, indent=4)

    logger.info(f'saved result in {file_name}')


def load_result(file_name: str) -> Result:
    """Load previous analysis result from file.

    This method is the reverse of
    [`save_result`](file_io.md#pymwp.file_io.save_result)
    and assumes the input matches the output of that method.

    Arguments:
        file_name: File to read.

    Raises:
        Exception: if `file_name` does not exist or cannot be read.

    Returns:
        Parsed result from file.
    """
    # read the file
    with open(file_name) as file_object:
        data = json.load(file_object)

    return Result.from_dict(**data)



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from __future__ import annotations

from typing import Union, Callable, Optional, Tuple, List

from . import B_TRIPLE
from .relation import SimpleRelation
from .semiring import UNIT_MWP, WEAK_MWP, POLY_MWP


class HonestPoly:
    """Models an honest polynomial.

    Attributes:
        variables (set[str]): List of variables.
        var_fmt(Callable): Variable formatter.
        op (str): Operator.
    """

    def __init__(self, operator: str, *init_vars: str):
        """Creates a honest polynomial.

        Arguments:
            operator (str): Operator.
            *init_vars (str): Arbitrary variables.
        """
        self.variables = set(init_vars)
        self.var_fmt: Optional[Callable[[str], str]] = None
        self.op = operator

    def __eq__(self, other: HonestPoly):
        return self.variables == other.variables

    @property
    def empty(self) -> bool:
        """True if honest polynomial is empty."""
        return len(self.variables) == 0

    @property
    def vars(self) -> List[str]:
        """Sorted list of variables."""
        var_list = [self.var_fmt(v) for v in self.variables] \
            if self.var_fmt else self.variables
        return sorted(list(var_list))

    @property
    def value(self) -> Union[int, str]:
        """String representation of HP value."""
        return 0 if self.empty else self.op.join(self.vars)

    def add(self, *identifier: str):
        """Add variables to HP."""
        for i in identifier:
            self.variables.add(i)

    def __str__(self):
        return str(self.value)


class MaxVar(HonestPoly):
    """Representation for m-variables."""

    def __init__(self, *init_vars: str):
        super().__init__(',', *init_vars)


class MwpBound:
    """Represents a mwp-bound.

    Attributes:
        x (Tuple[str]): List of m-variables.
        y (Tuple[str]): List of w-variables.
        z (Tuple[str]): List of p-variables.
    """

    def __init__(self, triple: str = None):
        """Create mwp-bound.

        Arguments:
            triple (str): Bound expression as bound_str.
        """
        x, y, z = self.parse(triple)
        self.x = MaxVar(*x)
        self.y = HonestPoly('+', *y)
        self.z = HonestPoly('*', *z)

    def __str__(self):
        return self.bound_poly(self)

    def __eq__(self, other: MwpBound):
        return (self.x == other.x
                and self.y == other.y
                and self.z == other.z)

    @property
    def bound_triple(self) -> B_TRIPLE:
        """Alternative bounds representation, as three tuples.

        Returns:
            Current bound as $(m_1,...m_n), (w_1,...w_n), (p_1,...p_n)$
                where the first contains list of variables in m,
                second contains variables in w, and last in p (if any).
        """
        return tuple(self.x.vars), tuple(self.y.vars), tuple(self.z.vars)

    @property
    def bound_str(self) -> str:
        """Alternative bounds representation, as a `;`-separated string.

        Returns:
            Current bound as $m;w;p$ where the first section contains
                list of variables in m, second contains variables in w,
                and last in p (if any). Multiple elements in the lists
                are separated by commas.
        """
        return f'{";".join([",".join(v) for v in self.bound_triple])}'

    def poly(self, k: str, compact: bool = False):
        bound_exp = MwpBound.bound_poly(self, compact)
        return f'{k}′{"≤" if compact else " ≤ "}{bound_exp}'

    @staticmethod
    def parse(value: str = None) -> B_TRIPLE:
        """Restore a bound from string-triple format."""
        return tuple([tuple(v.split(',')) if v else []
                      for v in value.split(";")]) if value \
            else (tuple(), tuple(), tuple())

    @staticmethod
    def bound_poly(mwp: MwpBound, compact=False):
        x, y, z, term = mwp.x, mwp.y, mwp.z, None
        # Any of the three variable lists might be empty
        if not x.empty and not y.empty:
            term = f'max({x},{y})'
        elif not x.empty:
            term = (f'max({x})' if len(x.vars) > 1 else str(x)) \
                if compact else (
                f'max({x},0)' if (len(x.vars) > 1 or not z.empty) else str(x))
        elif not y.empty:
            term = (f'max({y})' if len(y.vars) > 1 else str(y)) \
                if compact else (
                f'max({y},0)' if (len(y.vars) > 1 or not z.empty) else str(y))
        if term:
            return str(term) if z.empty else f'{term}+{z}'
        return str(z)

    def append(self, scalar: str, var_name: str):
        """Append variable dependency in the right list by scalar."""
        if scalar == UNIT_MWP:
            self.x.add(var_name)
        if scalar == WEAK_MWP:
            self.y.add(var_name)
        if scalar == POLY_MWP:
            self.z.add(var_name)


class Bound:
    """Represents a mwp-bound for a relation. If derivation succeeds,
    there is one mwp-bound expression for each input variable.

    Attributes:
        bound_dict (Dict[str, MwpBound]): Variable bounds.
    """

    LAND = '∧'

    def __init__(self, bounds: dict[str, str] = None):
        """Create a bound.

        Arguments:
            bounds (dict[str, str]): Dictionary of mwp-bounds.
        """

        self.bound_dict = dict([
            (k, MwpBound(triple=v)) for k, v in bounds.items()]) \
            if bounds else {}

    def __eq__(self, other: Bound):
        return (self.variables == other.variables and
                all(self.bound_dict[k] == other.bound_dict[k]
                    for k in self.variables))

    @property
    def variables(self) -> List[str]:
        """List of variables."""
        return list(self.bound_dict.keys())

    def calculate(self, relation: SimpleRelation) -> Bound:
        """Calculate bound from a simple-valued matrix.

        Arguments:
            relation: A simple-valued relation.

        Returns:
            The bound for the relation.
        """
        vars_, matrix = relation.variables, relation.matrix
        for col_id, name in enumerate(vars_):
            var_bound = MwpBound()
            for row_id in range(len(matrix)):
                var_bound.append(matrix[row_id][col_id], vars_[row_id])
            self.bound_dict[name] = var_bound
        return self

    def to_dict(self) -> dict:
        """A dictionary representation of a bound."""
        return dict([(k, v.bound_str)
                     for k, v in self.bound_dict.items()])

    def show(
            self, compact: bool = False, significant: bool = False,
            variables: Tuple[str] = None
    ) -> str:
        """Formatted display-string of a bound.

        Arguments:
            compact: reduce whitespace in the output
            significant: omit bounds that depend only on self
            variables: list of variables to display

        Returns:
            A formatted string of the bound.
        """
        key_filter = variables or list(self.bound_dict.keys())
        return f' {Bound.LAND} '.join([
            f'{k}′{"≤" if compact else " ≤ "}'
            f'{MwpBound.bound_poly(v, compact=compact)}'
            for k, v in self.bound_dict.items()
            if (not significant or str(k) != str(v)) and k in key_filter])



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
from collections import Counter
from functools import reduce
from itertools import product
from typing import Tuple, List, Set, Union, Optional, Generator, Callable

from . import SEQ, VECT, CHOICES

logger = logging.getLogger(__name__)


class Choices:
    """
    A compact representation of derivation choices.

    Attributes:
        valid (CHOICES): list of choice vectors.
        index (int): degree of choice.
    """

    def __init__(self, valid: CHOICES = None, index: int = -1):
        """Initialize representation from a precomputed vector.

        This is primarily useful for restoring a result from file.
        To create a choice representation, call
        [`generate()`](choice.md#pymwp.choice.Choices.generate) instead.
        """
        self.valid = valid or []
        if index < 0 and valid and len(valid):
            self.index = len(valid[0])
        else:
            self.index = index

    def __bool__(self):
        return not self.infinite

    @property
    def infinite(self):
        """True if no valid choice exists."""
        return len(self.valid) == 0 and self.index > 0

    @property
    def first(self) -> Optional[tuple[int]]:
        """Gets the first valid derivation choice, if at least one exists."""
        # take first vector, then first choice at each index
        return tuple([choices[0] for choices in self.valid[0]]) \
            if not self.infinite else None

    @property
    def n_bounds(self) -> int:
        """Number of bounds that can be generated from a choice vector.
        This is calculated directly from the vector form. It is the product
        of number of choices at each index: $\\prod_{c \\in v} |c|$.

        Example:
            1. Vector `[[0, 1, 2], [0, 1, 2], [0]]` allows making `[3, 3, 1]`
               choices/index. It has $3 * 3 * 1 = 9$ total bounds.

            2. A vector with choices/index: `[3, 1, 2, 1, 3, 3]` has
                $3^3 * 2^1 * 1^2 = 54$ possible bounds.

        Returns:
            Number of possible choices (non-distinct).
        """
        return sum([
            reduce(lambda total, n: total * n, lens, 1)
            for lens in [[len(x) for x in v] for v in self.valid]])

    def is_valid(self, *choices: int) -> bool:
        """Checks if sequence of choices can be made without infinity.

        Example:
           ```Python
           choice_obj.is_valid(0, 1, 2, 1, 1, 0)
           ```

        Arguments:
            choices: sequences of delta values to check.

        Returns:
            True if the given choices can be made without infinity.
        """
        for vector in self.valid:
            if len(choices) <= len(vector) and False not in \
                    [value in vector[idx] for idx, value in
                     enumerate(choices)]:
                return True
        return False

    def all(self) -> Generator[Tuple[int, ...]]:
        """Generator for all valid derivation choices.

        Note:
            Calling this method is very inefficient. Generally, we do not
            want to generate all choices. The method _yields_ the next value
            in series, so it is possible to step through the choices one at a
            time. Use with caution/only call if necessary.

        Yields:
            The next valid choice.
        """
        for choices in self.valid:
            for p in product(*choices):
                yield p

    @staticmethod
    def generate(domain: List[int], index: int, inf: Set[SEQ]) -> Choices:
        """Generate the choice representation.

        This works in two steps: (1) simplify delta sequences, (2) build
        choice vectors.

        Arguments:
            domain: valid choices at index, e.g. `[0, 1, 2]`.
            index: the length of the vector, e.g. 10. This is the same as
                number of assignments in the analyzed function.
            inf: set of deltas that lead to infinity.

        Returns:
            Generated choice object.
        """
        # simplify the set of infinities
        sequences = Choices.simplify(domain, inf)

        # now only min unique paths that lead to infinity remain
        # only sorted here for presentation purposes
        paths = [str(list(i)) for i in sorted(
            list(sequences), key=lambda x: (len(x), x))]
        logger.debug(f'infinity paths: {" # ".join(paths)}')

        # build vectors representing valid choices
        valid = Choices.build_choices(domain, index, sequences)
        return Choices(valid, index)

    @staticmethod
    def simplify(domain: List[int], sequences: Set[SEQ]) -> Set[SEQ]:
        """Generate the most simplified, equivalent representation of
        the set of choices that cause unwanted derivation results.

        Reduce sequences of deltas, until set of sequences cannot be reduced
        any further. Then remove all superset contained by some shorter
        sequence. Lastly, remove deltas that would generate an invalid
        choice vector. This process repeats until no more simplification
        can be applied.

        Arguments:
            domain: list of valid per index choices, e.g. `[0, 1, 2]`.
            sequences: set of delta sequences.

        Returns:
            Simplified list of infinity paths.
        """
        while True:
            while Choices.reduce(domain, sequences):
                continue
            while Choices.reduce_end(domain, sequences):
                continue
            len_before = len(sequences)
            sequences = Choices.unique_sequences(sequences)
            sequences = Choices.except_one(domain, sequences)
            len_after = len(sequences)
            if len_before == len_after or len_after == 0:
                return sequences

    @staticmethod
    def _reduce(
            domain: List[int], sequences: Set[SEQ],
            sub_eq: Callable[[SEQ, SEQ], bool], get_: Callable[[SEQ], int],
            keep_: Callable[[SEQ], SEQ]
    ) -> bool:
        """Implement sequence reduction from select direction.

        Arguments:
            domain: list of valid per index choices, e.g. `[0, 1, 2]`.
            sequences: set of delta sequences.
            sub_eq: subsequence comparison function.
            get_: choice value getter, e.g., first or last value of sequence.
            keep_: getter for sub-sequence that should be preserved.

         Returns:
            True if a reduction occurred and False otherwise.
        """
        for s1 in [s for s in sequences if len(s) > 1]:
            subs = [get_(s2) for s2 in sequences if sub_eq(s1, s2)]
            # all paths must exist
            if set(subs) == set(domain):
                # keep rest of sequence
                keep = keep_(s1)
                # remove all sequences contained by the shorter path
                Choices.remove_subset(keep, sequences)
                # finally add the shorter sequence to the set
                sequences.add(keep)
                return True
        return False

    @staticmethod
    def reduce(domain: List[int], sequences: Set[SEQ]) -> bool:
        """Look for first reducible sequence, if exists, then replace it.

        Example:
           We can reduce a sequences where deltas differ only on first value,
           never on index, and all possible choice values are represented
           in the first delta. Below, it does not matter which choice is
           made at index 0. The 3 paths can be collapsed into a single,
           shorter path: `(2,1)(1,4)`.

           ```
           (0,0) (2,1) (1,4)
           (1,0) (2,1) (1,4)
           (2,0) (2,1) (1,4)
           ```

        Arguments:
            domain: list of valid per index choices, e.g. `[0, 1, 2]`.
            sequences: set of delta sequences.

        Returns:
            True if a reduction occurred and False otherwise. The meaning of
                False is to say the operation is done and should not be
                repeated any further.
        """
        return Choices._reduce(
            domain, sequences, Choices.sub_equal,
            get_=lambda s2: s2[0][0], keep_=lambda s1: s1[1:])

    @staticmethod
    def reduce_end(domain: List[int], sequences: Set[SEQ]) -> bool:
        """Like `reduce`, but from end of sequence.

        Example:
           When deltas only differ at last index, and all choices
           occur at last index, reduce choices to a shorter path.
           E.g., Below, choice at index 5 is irrelevant; keep `(2,1) (1,4)`.

           ```
           (2,1) (1,4) (0,5)
           (2,1) (1,4) (1,5)
           (2,1) (1,4) (2,5)
           ```

        Arguments:
            domain: list of valid per index choices, e.g. `[0, 1, 2]`.
            sequences: set of delta sequences.

        Returns:
            True if a reduction occurred and False otherwise.
        """
        return Choices._reduce(
            domain, sequences, Choices.sub_equal_end,
            get_=lambda s2: s2[-1][0], keep_=lambda s1: s1[:-1])

    @staticmethod
    def except_one(domain: List[int], sequences: Set[SEQ]) -> Set[SEQ]:
        """Look for deltas of length=1 where all-but-one choices are made
        at same index. The remaining one choice can be eliminated from paths
        that contain it, because it would yield an invalid choice vector.

        Example:
            Given sequence `((0,0),), ((1,0),), ((2,0),(2,1),(1,4),)` and
            the goal to build a choice vector. Deltas `(0,0)` and `(1,0)` must
            be chosen. For `((2,0),(2,1),(1,4),)` one of the deltas must be
            chosen. But choosing `(0,0), (1,0), (2,0)` eliminates all choices
            at index 0. It suffices to keep sequence
            `((0,0),), ((1,0),), ((2,1),(1,4),)`.

        Arguments:
            domain: list of valid per index choices, e.g. `[0, 1, 2]`.
            sequences: set of delta sequences.

        Returns:
            A set of sequences after applying simplification.
        """
        l1 = [s[0] for s in sequences if len(s) == 1]
        while l1:
            (v, index) = l1.pop(0)
            matches = [itm for itm in l1 if itm[1] == index]
            values = [m[0] for m in matches]
            find = [(c, index) for c in domain
                    if c != v and c not in values]
            if len(find) == 1:
                for p in [s for s in sequences
                          if find[0] in s and len(s) > 1]:
                    sequences.add(tuple([x for x in p if x != find[0]]))
                    sequences.remove(p)
            map(l1.remove, matches)
        return sequences

    @staticmethod
    def unique_sequences(infinities: Set[SEQ]) -> Set[SEQ]:
        """Remove superset delta sequences.

        Arguments:
            infinities: set of delta sequences causing infinity.

        Returns:
            A list where all longer sequences, whose pattern is covered
                by some shorter sequence, are removed.
        """
        sequences = set()
        infinity_deltas: List[SEQ] = sorted(list(infinities), key=len)
        while infinity_deltas:
            first: SEQ = infinity_deltas.pop(0)
            Choices.remove_subset(first, infinity_deltas)
            sequences.add(first)
        return sequences

    @staticmethod
    def remove_subset(match: SEQ, items: Union[Set, List]):
        """If `match` is a subset of any item in `items`, removes the superset
        from items, in place.

        Arguments:
            match: single delta sequence.
            items: list of delta sequences to check against match.
        """
        for item in sorted(list(items)):
            if set(match).issubset(set(item)):
                items.remove(item)

    @staticmethod
    def sub_equal(first: SEQ, second: SEQ) -> bool:
        """Compare two delta sequences for equality, except their 0th value.

        Arguments:
            first: first delta sequence.
            second: second delta sequence.

        Returns:
            True if two delta sequences are equal excluding the 0th value,
                and False otherwise.
        """
        return first[0][1] == second[0][1] and first[1:] == second[1:]

    @staticmethod
    def sub_equal_end(first: SEQ, second: SEQ) -> bool:
        """Compare two delta sequences for equality, except their N-th
        value. This is like `sub_equal`, but comparison is done at the end of
        sequences.

        Arguments:
            first: first delta sequence.
            second: second delta sequence.

        Returns:
            True if two delta sequences are equal excluding the N-th
                value, and False otherwise.
        """
        return first[-1][1] == second[-1][1] and first[:-1] == second[:-1]

    @staticmethod
    def prod(values: list) -> int:
        """Compute the product of numeric list.

        Arguments:
            values: 1d list of numbers.

        Returns:
            Product of values.
        """
        return reduce((lambda x, y: x * y), values, 1)

    @staticmethod
    def build_choices(domain: List[int], index: int, infinities: Set[SEQ]) \
            -> CHOICES:
        """Build a list of distinct choice vectors excluding unwanted choices.

        This method works by taking a list of delta paths that give unwanted
        scalars (e.g., $\\infty$) and then negates those choices; the result
        is a list of choice vectors such that any remaining choice will give
        a valid derivation.

        Example:
           * Assume paths leading to $\\infty$ are:
             `[(0,0)], [(1,0)], [(1,1)(0,3)]`.
           * The choices that do not lead to $\\infty$ are:
             `[[[2], [0,2], [0,1,2]]` or `[[2], [0,1,2], [1,2]]]`.

        Arguments:
            domain: list of valid choices for one index, e.g. `[0, 1, 2]`.
            index: the length of the vector, e.g. 10.
            infinities: set of deltas that lead to unwanted choices.

        Returns:
            Choice vector that excludes all paths leading to unwanted choices.
        """
        if not infinities:
            return [[list(domain[:]) for _ in range(index)]]

        # noinspection PyTypeChecker
        # sort the infinity paths by length, the shortest first
        sorted_infty = sorted(list(infinities), key=len)

        # get length of each infinity path
        lens = [len(i) for i in sorted_infty]
        # helper for generating distinct combinations of indices
        iters = [Choices.prod(lens[idx + 1:]) for idx, _ in enumerate(lens)]

        # the product of path lengths gives the max number of distinct vectors
        max_ = Choices.prod(lens)
        logger.debug(f'maximum distinct vectors: {max_}')

        # noinspection PyTypeChecker
        # number of times each delta occurs in remaining paths
        delta_freq = Counter([j for sub in sorted_infty for j in sub])
        distinct = max(delta_freq.values()) == 1 if delta_freq else 0
        vectors = set()

        # generate all possible vectors by iterating the max count of
        # distinct vectors
        for iter_i in range(max_):

            # from iteration count, generate selectors for deltas
            # this one line does a lot - only edit if you have a good reason
            indices = [(iter_i // i) % x for x, i in zip(lens, iters)]

            # now choose the specific delta values, 1 value for each infinite
            # path, so that it is not possible to choose any bad path fully
            # this is same as taking cross product of deltas
            deltas = [sorted_infty[i][v] for i, v in enumerate(indices)]

            idx_freq = [i for v, i in set(deltas)]
            is_valid = all([idx_freq.count(n) < len(domain)
                            for n in set(idx_freq)])
            # This iteration will not produce a valid vector if all choices
            # are eliminated at some index.
            if not is_valid:
                continue

            # initialize a vector with all allowed choices
            vector = [set(domain[:]) for _ in range(index)]

            # iterate the infinity deltas to remove them from the vector
            for choice, idx in set(deltas):
                if choice in vector[idx]:
                    vector[idx].remove(choice)

            vector = tuple([tuple(entry) for entry in vector])
            # keep maximal and distinct choices
            if distinct or Choices.vect_new(vectors, vector):
                if not distinct:
                    Choices.vect_rm(vectors, vector)
                vectors.add(vector)

        # change the remaining choices at each index to lists (not sets)
        # so the vectors can be saved to file
        return Choices.to_choices(list(vectors))

    @staticmethod
    def to_choices(vectors: List[VECT]) -> CHOICES:
        """Convert a list of vectors to CHOICES type.
        This is a simple type conversion over the inner collection-types."""
        return [list([list(c) for c in v]) for v in vectors]

    @staticmethod
    def vect_new(vectors: Set[VECT], vector: VECT) -> bool:
        """Determines if a vector is distinct from all existing vectors.

        Arguments:
            vectors: a set of known vectors.
            vector: vector to check.

        Returns:
            True if vector does not occur in vectors.
        """
        return not next((Choices.vect_contains(v, vector)
                         for v in vectors), False)

    @staticmethod
    def vect_rm(vectors: Set[VECT], vector: VECT) -> None:
        """Remove from vectors those that are contained by vector.

        Arguments:
            vectors: a set of vectors.
            vector: vector compare against.
        """
        to_remove = [v for v in vectors if Choices.vect_contains(vector, v)]
        [vectors.remove(v) for v in to_remove]

    @staticmethod
    def vect_contains(a: VECT, b: VECT) -> bool:
        """Check if A allows making all choices of B.

        Arguments:
            a: first vector.
            b: second vector.

        Returns:
            True if A contains B, and False otherwise.
        """
        return all(all(ib in super_v for ib in sub)
                   for super_v, sub in zip(a, b))

    @staticmethod
    def vect_intersection(a: VECT, b: VECT) -> Optional[VECT]:
        """Intersection of two vectors.

        The intersection is the maximal choice at each index, permitted
        by both vectors A and B. The intersection is empty if, at some
        index, no choice exists.

        Example:
            Let a=`((0,1),(0,2))`, b=`((1,),(1,2))`, and c=`((1,),(1,))`.

            * The intersection of a&b is `((1,),(2,))`.
            * The intersection of a&c is None.
            * The intersection of b&c is `((1,),(1,))`.

        Arguments:
            a: first vector.
            b: second vector.

        Returns:
            A new vector that is the intersection of A and B; `None` when the
                intersection is empty.
        """
        tmp = tuple([tuple([ax for ax in av if ax in bv])
                     for av, bv in zip(a, b)])
        empty = next((True for x in tmp if len(x) == 0), False)
        return None if empty else tmp

    @staticmethod
    def intersection(c1: Choices, c2: Choices) -> Choices:
        """Intersection of two choice vectors.

        Intersection is a cross-product of non-empty vector intersections.
        If the none of the vectors intersect, the resulting Choice is empty.

        Arguments:
            c1: First Choice vector.
            c2: Second Choice vector.

        Returns:
              New Choices representing the intersection of the two arguments.
        """
        assert c1.index == c2.index
        choices = Choices.to_choices([j for sub in [
            [Choices.vect_intersection(v1, v2)
             for v2 in c2.valid] for v1 in c1.valid] for j in sub if j])
        return Choices(valid=choices, index=c1.index)

    @staticmethod
    def choice_reduce(*c: Choices) -> Choices:
        if len(c) == 0:
            return Choices()
        result = c[0]
        for n in range(1, len(c)):
            result = Choices.intersection(result, c[n])
        return result



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import logging
from typing import List, Tuple, Dict

from . import Coverage, Variables, FindLoops, COM_RES
from . import DeltaGraph, Polynomial, RelationList, Relation, Bound, Choices
from . import Result, FuncResult, FuncLoops, LoopResult, VResult
# noinspection PyPep8Naming
from .parser import Parser as pr
from .semiring import ZERO_MWP, UNIT_MWP, POLY_MWP, WEAK_MWP

logger = logging.getLogger(__name__)


class Analysis:
    """MWP analysis implementation."""

    DOMAIN = [0, 1, 2]

    @staticmethod
    def run(ast: pr.Node, res: Result = None, fin: bool = False,
            strict: bool = False, **kwargs) -> Result:
        """Run MWP analysis on AST.

        Arguments:
            ast (pr.Node): Parsed C source code AST Node.
            res (Result): Pre-initialized result object.
            fin (bool): Always run to completion.
            strict (bool): Require supported syntax.

        Returns:
            Analysis Result object.
        """
        result: Result = res or Result()
        logger.debug("started analysis")
        result.on_start()
        for f_node in [f for f in ast if pr.is_func(f)]:
            if Analysis.syntax_check(f_node, strict):
                func_res = Analysis.func(f_node, not fin)
                func_res.func_code = pr.to_c(f_node, True)
                result.add_relation(func_res)
        result.on_end().log_result()
        return result

    @staticmethod
    def func(node: pr.FuncDef, stop: bool) -> FuncResult:
        """Analyze a function.

        Arguments:
            node: parsed C source code function node
            stop: terminate if no bound exists

        Returns:
              Analysis result for provided function.
        """
        assert pr.is_func(node)
        name = node.decl.name
        logger.info(f"Analyzing {name}")
        result = FuncResult(name).on_start()

        # setup for function analysis
        variables, body = Variables(node).vars, (node.body.block_items or [])
        relations = RelationList.identity(variables=variables)
        total, num_v = len(body), len(variables)
        show_vars = ', '.join(variables) if num_v <= 5 else num_v
        logger.debug(f"{name} variables: {show_vars}")
        logger.debug(f"{total} top-level commands to analyze")

        # analyze body commands
        delta_infty, index = Analysis.cmds(relations, 0, body, stop)

        # evaluate choices + calculate a bound
        evaluated, choices, bound = False, None, None
        if not delta_infty:
            choices = relations.first.eval(Analysis.DOMAIN, index)
            if not choices.infinite:
                bound = Bound().calculate(
                    relations.first.apply_choice(*choices.first))
            evaluated = True
        # infinite by delta graph or by choice
        infinite = delta_infty or (evaluated and choices.infinite)

        # record results
        result.index = index
        result.infinite = infinite
        result.variables = relations.first.variables
        if not (infinite and stop):
            result.relation = relations.first
        if infinite and not stop:
            var_choices = relations.first.var_eval(Analysis.DOMAIN, index)
            result.inf_flows = relations.first.infty_pairs(
                [v for v, c in var_choices.items() if c.infinite])
        if not infinite:
            result.bound = bound
            result.choices = choices
        result.on_end()
        return result

    @staticmethod
    def cmds(relations: RelationList, index: int, nodes: List[pr.Node],
             stop: bool = True) -> Tuple[bool, int]:
        """Analyze some list of commands, typically body block statements.

        Arguments:
            relations (RelationList): Initialized relation list.
            index (int): Derivation index.
            nodes (List[pr.Node]): List of AST nodes to analyze.
            stop (bool): Set True to terminate early.

        Returns:
            True if nodes lead to infinity by delta graph.
        """
        if not nodes:
            return False, index
        delta_infty, total, dg = False, len(nodes), DeltaGraph()
        for i, node in enumerate(nodes):
            logger.debug(f'computing relation...{i} of {total}')
            index, rel_list, delta_infty_ = Analysis \
                .compute_relation(index, node, dg)
            delta_infty = delta_infty or delta_infty_  # cannot erase
            if stop and delta_infty:
                logger.debug('delta_graphs: infinite -> Exit now')
                break
            logger.debug(f'computing composition...{i} of {total}')
            relations.composition(rel_list)
        return delta_infty, index

    @staticmethod
    def syntax_check(node: pr.Node, strict: bool) -> bool:
        """Analyze function syntax and conditionally modify the AST.

        Arguments:
            node (pr.Node): An AST node.
            strict (bool): When true, AST will not be modified.

        Returns:
            True if analysis can be performed and False otherwise.
        """
        name = node.decl.name if pr.is_func(node) else 'node'
        cover = Coverage(node).report()
        if not cover.full and strict:
            logger.warning(f"{name} syntax is not fully analyzable")
            return False
        if not cover.full:
            cover.ast_mod()  # removes unsupported commands
            logger.warning(f"{name} syntax was modified")
        return True

    @staticmethod
    def compute_relation(index: int, node: pr.Node, dg: DeltaGraph) -> COM_RES:
        """Create a relation list corresponding for all possible matrices
        of an AST node.

        Arguments:
            index (int): Delta index.
            node (pr.Node): AST node to analyze.
            dg (DeltaGraph): DeltaGraph instance.

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        logger.debug("in compute_relation")
        if isinstance(node, (
                pr.Return, pr.Break, pr.Continue,
                pr.EmptyStatement, pr.Decl)):  # => skip
            return index, RelationList(), False
        if isinstance(node, pr.Assignment) and \
                isinstance(node.lvalue, pr.ID):
            rvalue = node.rvalue.expr if isinstance(
                node.rvalue, pr.Cast) else node.rvalue
            if isinstance(rvalue, pr.BinaryOp):
                return Analysis.binary_op(index, node)
            if isinstance(rvalue, pr.Constant):
                return Analysis.constant(index, node.lvalue.name)
            if isinstance(rvalue, pr.UnaryOp):
                return Analysis.unary_asgn(index, node, dg)
            if isinstance(rvalue, pr.ID):
                return Analysis.id(index, node)
        if isinstance(node, pr.UnaryOp):
            return Analysis.unary_op(index, node)
        if isinstance(node, pr.If):
            return Analysis.if_stmt(index, node, dg)
        if isinstance(node, (pr.While, pr.DoWhile)):
            return Analysis.while_loop(index, node, dg)
        if isinstance(node, pr.For):
            return Analysis.for_loop(index, node, dg)
        if isinstance(node, pr.Compound):
            return Analysis.compound(index, node, dg)
        if (isinstance(node, pr.FuncCall)
                and isinstance(node.name, pr.ID)
                and (node.name.name in ('assert', 'assume'))):
            return index, RelationList(), False

        Analysis._unsupported(pr.to_c(node))
        return index, RelationList(), False

    @staticmethod
    def id(index: int, node: pr.Assignment) -> COM_RES:
        """Analyze x = y, where data flows between two variables.

        Arguments:
            index: delta index
            node: AST node representing a simple assignment

        Returns:
            Updated index value, relation list, and an exit flag.
        """

        # ensure we have distinct variables on both sides of x = y
        if not isinstance(node.lvalue, pr.ID) \
                or isinstance(node.rvalue, pr.Constant) \
                or node.lvalue.name == node.rvalue.name:
            return index, RelationList(), False

        x, y = node.lvalue.name, node.rvalue.name
        vars_list = [[x], [y]]
        logger.debug(f'Computing relation {x} = {y}')

        # create a vector of polynomials based on operator type
        #     x   y
        # x | o   o
        # y | m   m    because x != y
        vector = [Polynomial('o'), Polynomial('m')]

        variables = vars_list[0] + vars_list[1]
        rel_list = RelationList.identity(variables)
        rel_list.replace_column(vector, x)
        return index, rel_list, False

    @staticmethod
    def binary_op(index: int, node: pr.Assignment) -> COM_RES:
        """Analyze binary operation, e.g. `x = y + z`.

        Arguments:
            index: delta index
            node: AST node representing a binary operation

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        logger.debug('Computing Relation: binary op')
        x = node.lvalue
        y = Analysis.rm_cast(node.rvalue.left)
        z = Analysis.rm_cast(node.rvalue.right)

        # operands cannot be unary, nested cast etc.
        assert isinstance(y, (pr.Constant, pr.ID))
        assert isinstance(z, (pr.Constant, pr.ID))

        non_constants = tuple([
            v.name if hasattr(v, 'name') else None
            for v in [x, y, z]])

        # create a vector of polynomials based on operator type
        index, vector = Analysis.create_vector(
            index, node.rvalue.op, non_constants)
        # build a list of unique variables but maintain order
        variables = list(dict.fromkeys(non_constants))
        # create relation list
        rel_list = RelationList.identity(variables)
        if hasattr(x, 'name'):
            rel_list.replace_column(vector, x.name)

        return index, rel_list, False

    @staticmethod
    def constant(index: int, variable_name: str) -> COM_RES:
        """Analyze a constant assignment of form `x = c` where x is some
        variable and c is constant.

        !!! info "From \"A Flow Calculus of mwp-Bounds...\""

            To deal with constants, just replace the program’s constants by
            variables and regard the replaced constants as input to these
            variables.

        Arguments:
            index: delta index
            variable_name: name of variable to which constant is being assigned

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        logger.debug('Constant value node')
        return index, RelationList([variable_name]), False

    @staticmethod
    def unary_asgn(index: int, node: pr.Assignment, dg: DeltaGraph) -> COM_RES:
        """Assignment where right-hand-size is a unary op e.g. `x = y++;`.

        Arguments:
            index (int): delta index
            node (pr.Assignment): Node with right-side unary operation.
            dg (DeltaGraph): DeltaGraph instance.

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        logger.debug('Computing Relation: unary')
        tgt, right, op = node.lvalue, node.rvalue.expr, node.rvalue.op
        init = f'{pr.to_c(node, True)} ==> '
        new_node = None

        if isinstance(right, pr.Constant):
            new_node = pr.Assignment('=', tgt, right)

        if isinstance(right, pr.ID):
            r_name = right.name
            if op in Coverage.INC_DEC:
                fst = Analysis.rewrite_id_inc_dec(node.rvalue)
                snd = pr.Assignment('=', tgt, pr.ID(r_name))
                cmds = [fst, snd] if op in Coverage.PREFIX else [snd, fst]
                new_node = pr.Compound(cmds)
            if op == Coverage.MINUS:  # flips sign
                neg_1 = pr.Constant('int', -1)
                r_node = pr.BinaryOp(Coverage.MULT, pr.ID(r_name), neg_1)
                new_node = pr.Assignment('=', tgt, r_node)
            if op == Coverage.PLUS:  # does nothing
                new_node = pr.Assignment('=', tgt, pr.ID(r_name))

        if op == Coverage.NEG:  # negation=> 0 or 1
            new_node = pr.Assignment('=', tgt, pr.Constant('int', 1))

        if op == Coverage.SIZEOF:  # var size in bytes ≤ 64
            new_node = pr.Assignment('=', tgt, pr.Constant('int', 64))

        if new_node:
            logger.debug(f'{init}{pr.to_c(new_node, True)}')
            return Analysis.compute_relation(index, new_node, dg)

        Analysis._unsupported(pr.to_c(node))
        return index, RelationList(), False

    @staticmethod
    def unary_op(index: int, node: pr.UnaryOp) -> COM_RES:
        """Analyze a standalone unary operation.

        Arguments:
            index (int): Delta index.
            node (pr.UnaryOp): AST node to analyze.

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        expr = Analysis.rm_cast(node.expr)
        if node.op in Coverage.INC_DEC and isinstance(expr, pr.ID):
            node_ = Analysis.rewrite_id_inc_dec(node)
            return Analysis.binary_op(index, node_)
        # could be recursive unary...
        # others operators do nothing without assignment.
        return index, RelationList(), False

    @staticmethod
    def rewrite_id_inc_dec(node: pr.UnaryOp):
        """Converts unary ++/-- operators to binary: x = x (op) 1."""
        expr = Analysis.rm_cast(node.expr)
        new_op, name = node.op[-1], expr.name
        const_1 = pr.Constant('int', 1)
        rvalue = pr.BinaryOp(new_op, pr.ID(name), const_1)
        node_ = pr.Assignment('=', pr.ID(name), rvalue)
        logger.debug(f'{pr.to_c(node)} rewrite to {pr.to_c(node_)}')
        return node_

    @staticmethod
    def rm_cast(node: pr.Node) -> pr.Node:
        """If Cast node, returns the expression of cast."""
        while isinstance(node, pr.Cast) and node.expr:
            node = node.expr
        return node

    @staticmethod
    def if_stmt(index: int, node: pr.If, dg: DeltaGraph) -> COM_RES:
        """Analyze an if statement.

        Arguments:
            index (int): Delta index.
            node (pr.If): if-statement AST node.
            dg (DeltaGraph): DeltaGraph instance.

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        logger.debug('computing relation (conditional case)')
        true_relation, false_relation = RelationList(), RelationList()

        index, exit_ = Analysis.if_branch(
            index, node.iftrue, true_relation, dg)
        if exit_:
            return index, true_relation, True
        index, exit_ = Analysis.if_branch(
            index, node.iffalse, false_relation, dg)
        if exit_:
            return index, false_relation, True

        relations = false_relation + true_relation
        return index, relations, False

    @staticmethod
    def if_branch(index: int, node: pr.If, relation_list: RelationList,
                  dg: DeltaGraph) -> Tuple[int, bool]:
        """Analyze `if` or `else` branch of a conditional statement.

        This method will analyze the body of the statement and update
        the provided relation. It can handle blocks with or without surrounding
        braces. It will return the updated index value.

        If branch does not exist (when else case is omitted) this
        method does nothing and returns the original index value without
        modification.

        Arguments:
            index (int): Current delta index value.
            node (pr.If): AST if statement branch node.
            relation_list (RelationList): Current relation list state.
            dg (DeltaGraph): DeltaGraph instance.

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        if node is not None:
            for child in (node.block_items or []) \
                    if hasattr(node, 'block_items') else [node]:
                index, rel_list, exit_ = Analysis \
                    .compute_relation(index, child, dg)
                if exit_:
                    return index, exit_
                relation_list.composition(rel_list)
        return index, False

    @staticmethod
    def while_loop(index: int, node: pr.While, dg: DeltaGraph) -> COM_RES:
        """Analyze a while loop.

        Arguments:
            index (int): Delta index.
            node (pr.while): While loop node.
            dg (DeltaGraph): DeltaGraph instance.

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        logger.debug("analysing while")

        relations = RelationList()
        for child in node.stmt.block_items \
                if hasattr(node, 'block_items') else [node.stmt]:
            index, rel_list, exit_ = Analysis.compute_relation(
                index, child, dg)
            if exit_:
                return index, rel_list, exit_
            relations.composition(rel_list)

        logger.debug('while loop fixpoint')
        relations.fixpoint()
        relations.while_correction(dg)
        dg.fusion()

        return index, relations, dg.is_empty

    @staticmethod
    def for_loop(index: int, node: pr.For, dg: DeltaGraph) -> COM_RES:
        """Analyze for loop node.

        Arguments:
            index (int): Delta index.
            node (pr.For): for loop node.
            dg (DeltaGraph): DeltaGraph instance.

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        comp, x_var = Coverage.loop_compat(node)
        if not comp:
            return index, RelationList(), False
        relations = RelationList(variables=[x_var])
        for child in node.stmt.block_items \
                if hasattr(node, 'block_items') else [node.stmt]:
            index, rel_list, exit_ = Analysis.compute_relation(
                index, child, dg)
            if exit_:
                return index, rel_list, True
            relations.composition(rel_list)

        logger.debug('loop fixpoint')
        relations.fixpoint()
        relations.loop_correction(x_var, dg)
        dg.fusion()
        return index, relations, dg.is_empty

    @staticmethod
    def compound(index: int, node: pr.Compound, dg: DeltaGraph) -> COM_RES:
        """Compound AST node contains zero or more children and is
        created by braces in source code.

        We analyze such compound node by recursively analysing its children.

        Arguments:
            index (int): Delta index.
            node (pr.Compound): Compound AST node.
            dg (DeltaGraph): DeltaGraph instance.

        Returns:
            Updated index value, relation list, and an exit flag.
        """
        relations = RelationList()

        if node.block_items:
            for node in node.block_items:
                index, rel_list, exit_ = Analysis.compute_relation(
                    index, node, dg)
                relations.composition(rel_list)
                if exit_:
                    return index, relations, True
        return index, relations, False

    @staticmethod
    def create_vector(index: int, op: str, variables: Tuple[str, ...]) \
            -> Tuple[int, List[Polynomial]]:
        """Build a polynomial vector based on operator and the operands
        of a binary operation statement that has form `x = y (operator) z`.

        For an AST node that represents a binary operation, this method
        generates a vector of polynomials based on the properties of that
        operation. The returned vector depends on the type of operator,
        how many operands are constants, and if the operands are equal.

        Arguments:
            index: delta index
            op: operator
            variables: variables in this operation `x = y (op) z` in order,
                where `y` or `z` is `None` if constant

        Returns:
             Updated index, list of Polynomial vectors
        """
        assert op in Coverage.BIN_OPS
        x, y, z = variables
        vector = []

        # when left variable does not occur on right side of assignment
        # x = … (if x not in …), i.e. when left side variable does not
        # occur on the right side of assignment, we prepend 0 to vector
        if x != y and x != z:
            vector.append(Polynomial(ZERO_MWP))

        if y is None or z is None:
            vector.append(Polynomial.from_scalars(
                index, UNIT_MWP, UNIT_MWP, UNIT_MWP))

        elif op == '*' and y == z:
            vector.append(Polynomial.from_scalars(
                index, WEAK_MWP, WEAK_MWP, WEAK_MWP))

        elif op == '*' and y != z:
            vector.append(Polynomial.from_scalars(
                index, WEAK_MWP, WEAK_MWP, WEAK_MWP))
            vector.append(Polynomial.from_scalars(
                index, WEAK_MWP, WEAK_MWP, WEAK_MWP))

        elif op in {'+', '-'} and y == z:
            vector.append(Polynomial.from_scalars(
                index, POLY_MWP, POLY_MWP, WEAK_MWP))

        elif op in {'+', '-'} and y != z:
            vector.append(Polynomial.from_scalars(
                index, UNIT_MWP, POLY_MWP, WEAK_MWP))
            vector.append(Polynomial.from_scalars(
                index, POLY_MWP, UNIT_MWP, WEAK_MWP))

        return index + 1, vector

    @staticmethod
    def _unsupported(command: any):
        """Keep for debugging extending parser+syntax support."""
        warning, endc = '\033[93m', '\033[0m'
        fmt_str = str(command or "").strip()
        logger.warning(f'{warning}Unsupported syntax {fmt_str}{endc}')


class LoopAnalysis(Analysis):
    """MWP analysis for loops."""

    @staticmethod
    def run(ast: pr.Node, res: Result = None, strict: bool = False, **kwargs) \
            -> Result:
        """Run loop analysis.

        Arguments:
            ast (pr.Node): Parsed C source code as an AST.
            res (Result): Pre-initialized result object.
            strict (bool): Require supported syntax.

        Returns:
            Analysis Result object.
        """
        result = res or Result()
        result.on_start()
        logger.debug("Starting loop analysis")
        functions = [f for f in ast if pr.is_func(f)]
        for func in functions:
            f_name = func.decl.name
            f_result = FuncLoops(f_name)
            f_result.on_start()
            logger.info(f"Analyzing {f_name}")
            # find loops and check/fix loop body syntax
            # nested loops are duplicated+lifted
            loops = [loop for loop in FindLoops(func).loops if
                     LoopAnalysis.syntax_check(loop, strict)]
            logger.debug(f"Total analyzable loops: {len(loops)}")
            # analyze each loop
            for loop in loops:
                f_result.loops.append(LoopAnalysis.inspect(loop))
            f_result.on_end()
            result.add_loop(f_result)
        result.on_end()
        result.log_result()
        return result

    @staticmethod
    def inspect(node: pr.LoopT) -> LoopResult:
        """Analyze a loop.

        Arguments:
            node (LOOP_T): A loop node.

        Returns:
            Loop analysis result.
        """
        assert pr.is_loop(node)
        result = LoopResult(pr.to_c(node)).on_start()

        # setup for loop analysis
        variables = Variables(node).vars
        relations = RelationList.identity(variables=variables)

        # analyze body commands, always run to completion
        infty, index = LoopAnalysis.cmds(relations, 0, [node], stop=False)

        # lame fix because of #148
        infty = infty or relations.first.eval(Analysis.DOMAIN, index).infinite

        # evaluate at variables
        result.variables = dict(zip(variables, map(
            lambda v: LoopAnalysis.get_result(
                relations.first, index, v), variables))) \
            if not infty else \
            LoopAnalysis.maybe_result(relations.first, index)

        result.on_end()
        return result

    @staticmethod
    def syntax_check(node: pr.LoopT, strict: bool) -> bool:
        """Analyze function syntax and conditionally modify the AST.

        Arguments:
            node (LOOP_T): An AST loop node.
            strict (bool): When true, AST will not be modified.

        Returns:
            True if analysis can be performed and False otherwise.
        """
        base = Analysis.syntax_check(node, strict)
        return base and pr.is_loop(node)

    @staticmethod
    def get_result(relation: Relation, index: int, v_name: str) -> VResult:
        """Find variable bounds when they are known to exist.

        Arguments:
            relation (Relation): Relation object.
            index (int): Degree of analysis choice.
            v_name (str): Name of variable to evaluate.

        Returns:
            Evaluation result for identified variable.
        """
        result = VResult(v_name)
        options = (('is_m', (WEAK_MWP, POLY_MWP)),
                   ('is_w', (POLY_MWP,)),
                   ('is_p', ()))
        # find the "least bound-choice": 0/m < has w < has p
        for attr, scalars in options:
            choices = relation.var_eval(
                Analysis.DOMAIN, index, v_name, *scalars)
            if not choices.infinite:
                setattr(result, attr, True)
                result.choices = choices
                break
        assert result.choices
        simple_mat = relation.apply_choice(*result.choices.first)
        result.bound = Bound().calculate(simple_mat).bound_dict[v_name]
        return result

    @staticmethod
    def maybe_result(relation: Relation, index: int) -> Dict[str, VResult]:
        """Evaluate variables when some variables are known to fail.

        Arguments:
            relation (Relation): Relation object.
            index (int): Degree of analysis choice.

        Returns:
            Dictionary of results, for each variable in relation.
        """
        variables = relation.variables
        p_bounds = dict(zip(variables, map(lambda v: relation.var_eval(
            Analysis.DOMAIN, index, v), variables)))
        fail = [v for v, c in p_bounds.items() if c.infinite]
        rest = dict([(v, p_bounds[v]) for v in (set(variables) - set(fail))])
        result = dict([(v, VResult(v)) for v in fail])
        fail_idx = [relation.variables.index(v) for v in fail]
        if rest:  # maybe some well-behaving variables?
            red = Choices.choice_reduce(*rest.values())
            assert not red.infinite  # should always give a choice?
            simple_mat = relation.apply_choice(*red.first)
            for v in rest.keys():
                idx = relation.variables.index(v)
                # assumption: if dep is 0, it is 0 in all derivations?
                deps = set([simple_mat.matrix[fi][idx] for fi in fail_idx])
                # only when variable has no dependency on failing ones
                result[v] = (LoopAnalysis.get_result(relation, index, v)
                             if deps == {ZERO_MWP} else VResult(v))
        return result



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# flake8: noqa: W605

from __future__ import annotations

from typing import List, Optional

from . import DeltaGraph, Relation


class RelationList:
    """
    Relation list holds a list of [`Relations`](relation.md).
    It provides methods for performing operations collectively on all
    relations in the list.

    Attributes:
        relations (List[Relation]): List of relations.
    """

    def __init__(self, variables: Optional[List[str]] = None,
                 relation_list: Optional[List[Relation]] = None):
        """To create a relations list, specify either `variables` or
        `relation_list`.

        Arguments:
            variables: List of variables used to initialize a relation on
                relation list.
            relation_list: List of relations for initializing relation list.

        Specifying a `relation_list` argument of a list of `Relations`,
        initializes a `RelationList` containing the provided relations.
        If no `relation_list` argument is provided, the constructor will
        create a relation list with one relation, using the provided
        `variables` to initialize that relation.

        Example:
            Create relation list using specific variables

            ```python
            rel_list = RelationList(['X0', 'X1', 'X2'])
            ```

            Generates a list with 1 relation:

            ```
            X0 |  +o  +o  +o
            X1 |  +o  +o  +o
            X2 |  +o  +o  +o
            ```

            Create relation list by providing relations

            ```python
            relations = [Relation(['X0', 'X1']), Relation(['X0'])]
            rel_list = RelationList(relation_list=relations)
            ```

            Generates a list with 2 relations:

            ```
            1:      X0  |  +o  +o
                    X1  |  +o  +o

            2:      X0  |  +o
            ```

            If no arguments are provided, the result is a relation list with an
            empty relation.

            ```python
            rel_list = RelationList()
            ```

            Generates a list with 1 empty relation.
        """
        self.relations = relation_list or [Relation(variables)]

    @staticmethod
    def identity(variables: List[str]) -> RelationList:
        """Create relation list that contains 1
        [identity relation](relation.md#pymwp.relation.Relation.identity).

        This is an alternative way to construct a relation list.

        Example:
            Create relation list containing an identity relation

            ```python
            rel_list = RelationList.identity(['X0', 'X1', 'X2'])
            ```

            Generates a list with 1 identity relation:

            ```
            X0 |  +m  +o  +o
            X1 |  +o  +m  +o
            X2 |  +o  +o  +m
            ```

        Arguments:
            variables: A list of variables.

        Returns:
            RelationList that contains identity relation generated using the
                provided variables.
        """
        return RelationList(relation_list=[Relation.identity(variables)])

    def __str__(self) -> str:
        relations = '\n\n'.join([f'{r}' for r in self.relations])
        divider = '\n' + ('-' * 72)
        return divider + '\n' + relations + divider

    def __add__(self, other):
        return RelationList(relation_list=[
            r1 + r2 for r1 in self.relations
            for r2 in other.relations])

    @property
    def first(self) -> Relation:
        """Gets the first relation in relation list."""
        return self.relations[0]

    def replace_column(self, vector: list, variable: str) -> None:
        """For each relation in a relation list, replace column with a provided
         vector, in place.

        This method takes as input a variable, then finds the index of that
        variable based on its name, then applies the column replacement at the
        discovered index.

        Arguments:
            vector: vector that will replace a column
            variable: variable value; replace will occur at the index of this
                variable.

        Raises:
            ValueError: if variable does not exists some relation belonging
                to this relation list.
        """

        self.relations = [rel.replace_column(vector, variable)
                          for rel in self.relations]

    def composition(self, other: RelationList) -> None:
        """Apply composition to all relations in two relation lists.

        This method takes as argument `other` relation list, then composes the
        product of `self` and `other` by computing the product of each
        relation, for all combinations.

        Composition occurs in place. After composition `self` will contain all
        unique relations obtained during composition.

        To compose `RelationList` and a single `Relation`, see
        [`one_composition()`](relation_list.md#pymwp.relation_list.
        RelationList.one_composition).

        Arguments:
            other: RelationList to compose with `self`
        """
        new_list = []
        for r1 in self.relations:
            for r2 in other.relations:
                output = r1 * r2
                if not RelationList.contains_matrix(new_list, output.matrix):
                    new_list.append(output)

        self.relations = new_list

    @staticmethod
    def contains_matrix(search_in: List[Relation], matrix: List[List]) -> bool:
        """Check if a list of relations contains the provided matrix.

        Arguments:
            search_in: list of relations to search
            matrix: search value to look for

        Returns:
            `True` if matrix is found somewhere in the list of relations and
                `False` otherwise.
        """
        for relation in search_in:
            if relation.matrix == matrix:
                return True
        return False

    def one_composition(self, relation: Relation) -> None:
        """Compose each relation in a relation list × relation.

        This method iterates current relation list and applies
        [`composition()`](relation.md#pymwp.relation.Relation.composition) to
        each of its relations, using argument `relation` as the other operand.

        Arguments:
            relation: relation to compose with relations in current list.
        """
        self.relations = [rel * relation for rel in self.relations]

    def fixpoint(self) -> None:
        """Apply [fixpoint](relation.md#pymwp.relation.Relation.fixpoint)
         to all relations in relation list."""
        self.relations = [rel.fixpoint() for rel in self.relations]

    def show(self) -> None:
        """Display relation list."""
        print(str(self))

    def while_correction(self, dg: DeltaGraph) -> None:
        """Apply [`while_correction()`](relation.md#pymwp.relation.Relation
        .while_correction) to all relations in a relation list."""
        for rel in self.relations:
            rel.while_correction(dg)

    def loop_correction(self, x_var: str, dg: DeltaGraph) -> None:
        """Apply loop correction."""
        for rel in self.relations:
            rel.loop_correction(x_var, dg)



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# flake8: noqa: W605

from typing import List

ZERO_MWP: str = "o"
"""Scalar that represents no dependency (0) in the analysis (`'o'`)."""

UNIT_MWP: str = "m"
"""Scalar that represents maximally linear flow in the analysis (`'m'`)."""

WEAK_MWP: str = "w"
"""Scalar that represents weak polynomial flow in the analysis (`'w'`)."""

POLY_MWP: str = "p"
"""Scalar that represents a polynomial flow in the analysis (`'p'`)."""

INFTY_MWP: str = "i"
"""Scalar that represents failure in the analysis (`'i'`), $\\infty$."""

KEYS: List[str] = [ZERO_MWP, UNIT_MWP, WEAK_MWP, POLY_MWP, INFTY_MWP]
"""Different scalar values: `"o", "m", "w", "p", "i"`"""

__DICT_PROD: dict = {
    "o": {"o": "o", "m": "o", "w": "o", "p": "o", "i": "i"},

    "m": {"o": "o", "m": "m", "w": "w", "p": "p", "i": "i"},

    "w": {"o": "o", "m": "w", "w": "w", "p": "p", "i": "i"},

    "p": {"o": "o", "m": "p", "w": "p", "p": "p", "i": "i"},

    "i": {"o": "i", "m": "i", "w": "i", "p": "i", "i": "i"}
}

__DICT_SUM: dict = {
    "o": {"o": "o", "m": "m", "w": "w", "p": "p", "i": "i"},

    "m": {"o": "m", "m": "m", "w": "w", "p": "p", "i": "i"},

    "w": {"o": "w", "m": "w", "w": "w", "p": "p", "i": "i"},

    "p": {"o": "p", "m": "p", "w": "p", "p": "p", "i": "i"},

    "i": {"o": "i", "m": "i", "w": "i", "p": "i", "i": "i"}
}


def mwp_sort(scalars: List[str]):
    """Ascending sort of scalars (o < m < w < p < $\\infty$)."""
    return sorted(scalars, key=lambda x: KEYS.index(x))


def prod_mwp(scalar1: str, scalar2: str) -> str:
    """Compute product of two scalars.

    | $\\times$  | $0$       | $m$       | $w$        | $p$       | $\\infty$ |
    | ---        | ---       | ---       | ---        | ---       | --- |
    | $0$        | $0$       | $0$       | $0$        | $0$       | $\\infty$ |
    | $m$        | $0$       | $m$       | $w$        | $p$       | $\\infty$ |
    | $w$        | $0$       | $w$       | $w$        | $p$       | $\\infty$ |
    | $p$        | $0$       | $p$       | $p$        | $p$       | $\\infty$ |
    | $\\infty$  | $\\infty$ | $\\infty$ | $\\infty$  | $\\infty$ | $\\infty$ |

    Arguments:
        scalar1: scalar value.
        scalar2: scalar value.

    Raises:
        Exception: if `scalar1` or `scalar2` is not in KEYS.

    Returns:
        Product of scalar1 * scalar2.
    """
    if scalar1 in KEYS and scalar2 in KEYS:
        return __DICT_PROD[scalar1][scalar2]
    else:
        raise Exception(
            f"trying to use {scalar1} and {scalar2} as keys for prod…")


def sum_mwp(scalar1: str, scalar2: str) -> str:
    """Compute sum of two scalars.

    | $+$        | $0$       | $m$       | $w$        | $p$       | $\\infty$ |
    | ---        | ---       | ---       | ---        | ---       | --- |
    | $0$        | $0$       | $m$       | $w$        | $p$       | $\\infty$ |
    | $m$        | $m$       | $m$       | $w$        | $p$       | $\\infty$ |
    | $w$        | $w$       | $w$       | $w$        | $p$       | $\\infty$ |
    | $p$        | $p$       | $p$       | $p$        | $p$       | $\\infty$ |
    | $\\infty$  | $\\infty$ | $\\infty$ | $\\infty$  | $\\infty$ | $\\infty$ |

    Arguments:
        scalar1: scalar value.
        scalar2: scalar value.

    Raises:
        Exception: if `scalar1` or `scalar2` is not in KEYS.

    Returns:
        Sum of scalar1 + scalar2.
    """
    if scalar1 in KEYS and scalar2 in KEYS:
        return __DICT_SUM[scalar1][scalar2]
    else:
        raise Exception(
            f"trying to use {scalar1} and {scalar2} as keys for sum…")
